from typing import Any
from typing import Dict
from typing import List
from typing import Set

KEYWORDS = ["exclude", "__exclude__"]


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
    is_params: bool,
) -> List[str]:
    """
    Resolve any exclude statements

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    :param is_params: True for parameter sources
    """
    if "exclude" not in state:
        return []
    exc = state.pop("exclude", [])
    if not isinstance(exc, List):
        hub.idem.RUNS[name]["errors"].append(
            f"Exclude Declaration in SLS {sls_ref} is not formed as a list"
        )
    if exc:
        state.setdefault("__exclude__", []).extend(exc)

    return []



import pprint
from typing import Any
from typing import Dict
from typing import List


def display(hub, data):
    """
    Transform sls data into a format that is easily used in jmespath searches
    """
    return pprint.pformat(hub.output.jmespath.prepare(data))


def prepare(hub, data: Dict[str, Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Transform sls data into a format that is easily used in jmespath searches
    """
    if isinstance(data, List):
        # It's already in jmespath format
        return data
    ret = []
    if not data:
        return ret
    for sls_name, sls_data in data.items():
        for sls_ref, sls_params in sls_data.items():
            ret.append(
                {
                    "name": sls_name,
                    "ref": sls_ref,
                    "resource": sls_params,
                }
            )
    return ret


def revert(hub, obj: List[Dict[str, Any]]) -> Dict[str, Dict[str, Any]]:
    """ "
    Transform jmespath data back into an sls style
    """
    ret = {}
    for item in obj:
        ret[item["name"]] = {item["ref"]: item["resource"]}
    return ret



import copy
import fnmatch
import inspect
import warnings
from typing import Any
from typing import Dict
from typing import List


def gen_chunk_func_tag(hub, chunk):
    """
    Generate the unique tag used to track the execution of the chunk
    """
    return f'{_gen_tag(chunk)}{chunk["fun"]}'


def gen_chunk_esm_tag(hub, chunk: Dict[str, str]) -> str:
    """
    Generate the unique tag used to track the execution of the chunk in the esm
    """
    return _gen_tag(chunk)


def _gen_tag(chunk: Dict[str, str]) -> str:
    """
    Generate the unique tag used to track the execution of the chunk

    This will be used at a module level, agnostic to function names
    """
    tag = f'{chunk["state"]}_|-{chunk["__id__"]}_|-{chunk["name"]}_|-'

    # if `name_prefix` is present and `name` is formed with `name_prefix`, then create tag based on `name_prefix`
    if chunk.get("name_prefix") and chunk.get("name_prefix") in chunk["name"]:
        tag = f'{chunk["state"]}_|-{chunk["__id__"]}_|-{chunk["name_prefix"]}_|-'

    return tag


def get_chunks(hub, low, state, name):
    """
    Search in the low state for the chunk with the given designation
    """
    rets = []
    for chunk in low:
        if state == "sls":
            if fnmatch.fnmatch(chunk["__sls__"], name):
                rets.append(chunk)
                continue
        if state == chunk["state"]:
            if fnmatch.fnmatch(chunk["name"], name) or fnmatch.fnmatch(
                chunk["__id__"], name
            ):
                rets.append(chunk)
    return rets


def find_name(hub, name, state, high):
    """
    Scan high data for the id referencing the given name and return a list of (IDs, state) tuples that match
    Note: if `state` is sls, then we are looking for all IDs that match the given SLS
    """
    ext_id = []
    if name in high:
        ext_id.append((name, state))
    # if we are requiring an entire SLS, then we need to add ourselves to everything in that SLS
    elif state == "sls":
        for nid, item in high.items():
            if item["__sls__"] == name:
                ext_id.append((nid, next(iter(item))))
    # otherwise we are requiring a single state, lets find it
    else:
        # We need to scan for the name
        for nid in high:
            if state in high[nid]:
                if isinstance(high[nid][state], list):
                    for arg in high[nid][state]:
                        if not isinstance(arg, dict):
                            continue
                        if len(arg) != 1:
                            continue
                        if arg[next(iter(arg))] == name:
                            ext_id.append((nid, state))
    return ext_id


def format_call(
    hub,
    fun,
    data,
    ignore_changes: List = None,
    expected_extra_kws=(),
    enforced_state: Dict[str, Any] = None,
):
    """
    Build the required arguments and keyword arguments required for the passed
    function.
    :param fun: The function to get the argspec from
    :param data: A dictionary containing the required data to build the
                 arguments and keyword arguments.
    :param ignore_changes: A list of path of parameters that will be assigned to None value
                           to ignore being updated in present().
    :param expected_extra_kws: Any expected extra keyword argument names which
                               should not trigger a :ref:`SaltInvocationError`
    :param enforced_state: A dictionary with the parameters from a previous run
    :returns: Function keyword arguments.
    """
    enforced_state = enforced_state or {}

    args = []
    kwargs = {}
    keywords = False

    sig = fun.signature
    for name, param in sig.parameters.items():
        if name == "hub":
            continue
        elif param.kind.name == "POSITIONAL_OR_KEYWORD":
            if param.default is inspect._empty:
                if name in enforced_state:
                    # get it from enforced state before state
                    kwargs[name] = enforced_state[name]
                else:
                    # Warn about required argument that is not stored in ESM
                    if enforced_state and name != "ctx":
                        msg = (
                            f"Function {fun.ref}.{fun.func.__name__} argument '{name}' is required, "
                            f"but is not found in ESM cache for {gen_chunk_func_tag(hub, data)}"
                        )
                        warnings.warn(msg, RuntimeWarning)
                        hub.log.warning(msg)
                    args.append(name)
            else:
                kwargs[name] = enforced_state.get(name, param.default)
                # Validate a boolean value for arguments of type boolean
                if isinstance(param.default, bool):
                    if data.get(name) and not isinstance(data.get(name), bool):
                        raise TypeError(
                            f"{fun.ref}.{fun.func.__name__} is expecting a boolean value for '{name}' but got '{data.get(name)}'"
                        )
        elif param.kind.name == "KEYWORD_ONLY":
            # get it from enforced state before default
            kwargs[name] = enforced_state.get(name, param.default)
        elif param.kind.name == "VAR_KEYWORD":
            keywords = True

    # Since we WILL be changing the data dictionary, let's change a copy of it
    data = data.copy()
    is_existing_resource = bool(enforced_state) or bool(data.get("resource_id"))

    missing_args = []

    for key in kwargs:
        try:
            # Do not override enforced_state value with None,
            # unless no value is set
            val = data.pop(key)
            if val is not None or kwargs[key] is inspect._empty:
                kwargs[key] = val

            # if the execution flow is for recreating a resource,
            # then we should override enforced_state value with None for resource_id
            if key == "resource_id" and data.get("recreation_flow", False):
                kwargs[key] = val
        except KeyError:
            # Let's leave the default value in place
            pass

    # If ignore_changes contains parameters, then we try to assign those parameters with None to skip updating.
    # For resource recreation flow, we should not nullify the parameters in ignore_changes.
    if (
        ignore_changes
        and is_existing_resource
        and not data.get("recreation_flow", False)
    ):
        hub.idem.tool.ignore_changes.ignore_parameter_changes(
            ignore_changes=ignore_changes,
            params=kwargs,
            param_signatures=sig.parameters,
        )

    while args:
        # Every function takes ctx as an argument.
        # All others are required function arguments that do not exist in ESM
        arg = args.pop(0)
        try:
            kwargs[arg] = data.pop(arg)
        except KeyError:
            missing_args.append(arg)

    if missing_args:
        raise ValueError(
            f"{fun.ref}.{fun.func.__name__} is missing required argument(s): {', '.join(missing_args)}"
        )

    if keywords:
        # The function accepts **kwargs, any non expected extra keyword
        # arguments will made available.
        for key, value in data.items():
            if key in expected_extra_kws:
                continue
            kwargs[key] = value

        # No need to check for extra keyword arguments since they are all
        # **kwargs now. Return
        return kwargs

    # Did not return yet? Lets gather any remaining and unexpected keyword
    # arguments
    extra = {}
    for key, value in data.items():
        if key in expected_extra_kws:
            continue
        extra[key] = copy.deepcopy(value)

    if extra:
        # Found unexpected keyword arguments, raise an error to the user
        if len(extra) == 1:
            msg = "'{0[0]}' is an invalid keyword argument for '{1}'".format(
                list(extra.keys()), f"{fun.__module__}.{fun.__name__}"
            )
        else:
            msg = "{} and '{}' are invalid keyword arguments for '{}'".format(
                ", ".join([f"'{e}'" for e in extra][:-1]),
                list(extra.keys())[-1],
                f"{fun.__module__}.{fun.__name__}",
            )
        warnings.warn(msg, RuntimeWarning)
        hub.log.warning(msg)
    return kwargs


def ishashable(hub, obj):
    """
    A simple test to verify if a given object is hashable and can therefore
    be used as a key in a dict
    """
    try:
        hash(obj)
    except TypeError:
        return False
    return True


def iter_high(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield id_, body, state, run, arg


def iter_high_leaf_args(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield from _iter_arg_chain(id_, body, state, run, "", arg)


def _iter_arg_chain(id_, body, state, run, arg_key_def, arg):
    if isinstance(arg, dict):
        for arg_key, arg_value in arg.items():
            # Escape dictionary references in the key definition to skip evaluating by arg_bind resolver.
            arg_key = arg_key.replace("[", "[\\")
            if arg_key_def:
                current_arg_key_def = arg_key_def + ":" + arg_key
            else:
                current_arg_key_def = arg_key

            if isinstance(arg_value, dict):
                yield from _iter_arg_chain(
                    id_, body, state, run, current_arg_key_def, arg_value
                )
            elif isinstance(arg_value, list):
                for index, arg_value_item in enumerate(arg_value):
                    yield from _iter_arg_chain(
                        id_,
                        body,
                        state,
                        run,
                        current_arg_key_def + f"[{index}]",
                        arg_value_item,
                    )
            else:
                yield id_, body, state, run, current_arg_key_def, arg_value
    elif isinstance(arg, list):
        for index, arg_value_item in enumerate(arg):
            yield from _iter_arg_chain(
                id_, body, state, run, arg_key_def + f"[{index}]", arg_value_item
            )
    else:
        yield id_, body, state, run, arg_key_def, arg


def get_enforced_state(hub, chunk, managed_state):
    tag = hub.idem.tools.gen_chunk_func_tag(chunk)
    esm_tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
    new_chunk = {**chunk, "__id__": f"{chunk['__id__']}_create_new"}
    new_tag = hub.idem.tools.gen_chunk_func_tag(new_chunk)
    new_esm_tag = hub.idem.tools.gen_chunk_esm_tag(new_chunk)

    enforced_state = managed_state.get(new_esm_tag) or managed_state.get(new_tag)

    if not enforced_state:
        enforced_state = managed_state.get(esm_tag) or managed_state.get(tag)

    return enforced_state



from typing import Any
from typing import Dict


def define(hub) -> Dict[str, Any]:
    """
    Define how the watch requisite should behave
    """
    return {
        "result": True,
        "changes_post": "mod_watch",
    }



import json
from typing import Any
from typing import Dict


async def put(
    hub,
    body: Any,
    profile: str = "default",
    tags: Dict[str, Any] = None,
):
    if tags is None:
        tags = {}

    await hub.evbus.broker.put(
        profile=profile, body=dict(tags=tags, message=body, run_name=hub.idem.RUN_NAME)
    )


def put_nowait(
    hub,
    body: Any,
    profile: str = "default",
    tags: Dict[str, Any] = None,
):
    if tags is None:
        tags = {}

    hub.evbus.broker.put_nowait(
        profile=profile, body=dict(tags=tags, message=body, run_name=hub.idem.RUN_NAME)
    )


async def publish_error_callback_handler(hub, error_data: dict):
    """
    A default idem callback handler for publish event failure which pushes an error event back to the queue

    .. code-block:: json

        {
            "tags": {
                "plugin": "<plugin-name>",
                <original-event-tags>
            },
            "publish_event_error_message": "<error-message-from-publish-failure>",
            "run_name": "<run-name>"
        }

    """
    if error_data:
        original_body = json.loads(error_data["body"].decode())
        new_tags = {
            "ref": "idem.event.publish_error_callback_handler",
            "plugin": error_data.get("plugin", None),
        }
        run_name = ""
        if original_body:
            if original_body.get("publish_event_error_message"):
                # If this event failed to publish error event, return it.
                # The error would be logged already with evbus
                return

            original_tags = original_body.get("tags", {})
            new_tags = original_tags | new_tags
            run_name = original_body.get("run_name", None)

        hub.evbus.broker.put_nowait(
            profile=error_data["profile"],
            body=dict(
                tags=new_tags,
                run_name=run_name,
                publish_event_error_message=str(error_data["exception"]),
            ),
        )



# States for exec modules that implement the "auto_state" contract
import uuid

import dict_tools.differ as differ
import pop.contract

__contracts__ = ["resource"]


async def present(hub, ctx, name: str, exec_mod_ref=None, **kwargs):
    """
    Create a resource if it doesn't exist, update it if it does exist
    """
    result = dict(comment="", old_state={}, new_state={}, name=name, result=True)
    before = await hub.pop.loop.wrap(hub.exec[exec_mod_ref].get, ctx, name, **kwargs)

    if not before:
        if ctx.test:
            result["comment"] = f"Would create {exec_mod_ref}:{name}"
            return result
        result["result"] = (
            await hub.pop.loop.wrap(hub.exec[exec_mod_ref].create, ctx, name, **kwargs)
        ).result
        if result["result"]:
            result["comment"] = f"Created '{exec_mod_ref}:{name}'"
        else:
            # There was an error in creation
            result["comment"] = f"Could not create '{exec_mod_ref}:{name}'"
            return result
    else:
        result["comment"] = f"'{exec_mod_ref}:{name}' already exists"
        if ctx.test:
            result["comment"] = f"Would update {exec_mod_ref}:{name}"
            return result

        # Update the resource
        result["result"] = (
            await hub.pop.loop.wrap(hub.exec[exec_mod_ref].update, ctx, name, **kwargs)
        ).result

    after = await hub.pop.loop.wrap(hub.exec[exec_mod_ref].get, ctx, name, **kwargs)
    result["changes"] = differ.deep_diff(
        before.ret if before.ret else dict(), after.ret if after.ret else dict()
    )
    if before.ret and result["changes"]:
        result["comment"] = f"Updated '{exec_mod_ref}:{name}'"

    result["old_state"] = before.ret
    result["new_state"] = after.ret
    return result


async def absent(hub, ctx, name: str, exec_mod_ref=None, **kwargs):
    """
    Remove a resource if it exists
    """
    result = dict(comment="", old_state={}, new_state={}, name=name, result=True)
    before = await hub.pop.loop.wrap(hub.exec[exec_mod_ref].get, ctx, name, **kwargs)

    if before:
        if ctx.test:
            result["comment"] = f"Would delete {exec_mod_ref}:{name}"
            return result

        await hub.pop.loop.wrap(hub.exec[exec_mod_ref].delete, ctx, name, **kwargs)
        result["comment"] = f"Deleted '{exec_mod_ref}:{name}'"
    else:
        result["comment"] = f"'{exec_mod_ref}:{name}' already absent"
        return result

    after = await hub.pop.loop.wrap(hub.exec[exec_mod_ref].get, ctx, name, **kwargs)

    result["old_state"] = before.ret
    result["new_state"] = after.ret
    return result


async def describe(hub, ctx):
    """
    Create "present" states for a resource based on an "auto_state" exec module plugin
    """
    exec_mod_ref = ctx.exec_mod_ref
    result = {}

    ret = await hub.pop.loop.unwrap(hub.exec[exec_mod_ref].list(ctx))

    if not (ret["result"] and ret["ret"]):
        return result

    for name, resource in ret["ret"].items():
        create_contract: pop.contract.Contracted = hub.exec[exec_mod_ref].create
        present_state = [{"name": name}]

        # Add the keyword arguments in the proper order
        for param, inspection in create_contract.signature.parameters.items():
            if param in ("hub", "ctx", "name", "kwargs"):
                # Skip over the ones we know about
                continue
            if param in resource:
                present_state.append({param: resource.pop(param)})
            else:
                present_state.append({param: inspection.default})

        result[f"{name}-{uuid.uuid4()}"] = {f"{exec_mod_ref}.present": present_state}

    return result



import os.path
from typing import List


REQUISITES_TO_UPDATE_FOR_SLS_RUN = [
    "require",
    "require_in",
    "require_any",
    "listen",
    "arg_bind",
]


async def run(
    hub,
    ctx,
    name: str,
    sls_sources: List[str],
    params: List[str] = None,
    **kwargs,
):
    """
    State module to structure a group of SLS files into an independent construct that
    can be invoked from SLS code multiple times with different set of arguments and parameters

    Args:
        name(string):
            An idem name of the resource.

        sls_sources(list):
            List of sls files to run

        params(list, Optional):
            List of params files. All the params provided in these files will be consolidated and these consolidated params
            will override the params passed in Idem run. These consolidated params along with params passed to Idem run
            will be used to run the sls files provided in sls_sources. Defaults to None.

        kwargs(Dict[str, Any], Optional):
            parameters passed in kwargs are used as parameters to resolve parameters in sls_sources files.

        Request Syntax:
           .. code-block:: sls

              [sls-run-name]:
                sls.run:
                  - sls_sources:
                      - 'string'
                  - params:
                      - 'string'

        Returns:
            Dict[str, Any]

        Examples:
           .. code-block:: sls

               service4:
                sls.run:
                  - sls_sources:
                      - sls.service_file1
                      - sls.service_file2
                  - params:
                      - params.file1
    """

    result = dict(comment=[], name=name, old_state=None, new_state=None, result=True)

    # The sls_sources and params provided to this state are paths to files.
    # To resolve the location of files and parse them we use this hub.OPT.idem.tree which gives us the base path.
    sls_sources_path = list()
    param_sources_path = list()
    if hub.OPT.idem.tree:
        tree = f"file://{hub.OPT.idem.tree}"
        hub.log.debug(f"Added tree to sls and param sources: {tree}")
        sls_sources_path.append(tree)
        param_sources_path.append(tree)

    main_run_name = ctx.run_name
    # Create a temporary run name to use while compiling the sls_sources passed to this state
    temporary_run_name = main_run_name + "." + name

    # Create a temporary run to process the new sls block
    await hub.idem.state.create(
        name=temporary_run_name,
        sls_sources=sls_sources_path,
        # Allow a different render pipe to be used for the new render block
        # Default to the renderer of the main run
        render=hub.idem.RUNS[main_run_name].get("render"),
        # Copy state.apply parameters from the main run
        **{
            k: hub.idem.RUNS[main_run_name].get(k)
            for k in (
                "runtime",
                "subs",
                "cache_dir",
                "test",
                "acct_file",
                "acct_key",
                "acct_profile",
                "acct_blob",
                "managed_state",
                "param_sources",
                "acct_data",
                "invert_state",
            )
        },
    )

    sls_source_base_path = hub.OPT.idem.sls
    # if sls_sources_path is present in kwargs use that path as base path for resolving sls_sources
    # This indicates that this sls.run is a nested sls.run and sls_sources should be resolved
    # relative to parent sls.run path
    if "sls_sources_path" in kwargs:
        sls_source_base_path = kwargs["sls_sources_path"]

    # parse the sls_sources provided to this state with the consolidated params and get high data
    src = hub.idem.sls_source.init.get_refs(
        sources=sls_sources_path, refs=sls_source_base_path
    )

    # Gather params files provided to this file and combine with idem run params
    # we gather params from idem run, params files provided to this state and inline params provided to this state.
    # will combine the params from all the three sources. If there are common params, Inline params takes precedence
    # followed by params files provided to this state.
    run_params_ret = await _gather_params(
        hub, main_run_name, temporary_run_name, params, src["sls_sources"], kwargs
    )
    if "errors" in run_params_ret:
        result["comment"] = [
            f"Error in gathering params files for sls.run {name}"
        ] + run_params_ret["errors"]
        result["result"] = False
        hub.idem.RUNS.pop(temporary_run_name, None)
        return result

    # Resolve the new sls_sources with the main run
    gather_data = await hub.idem.resolve.init.gather(
        temporary_run_name,
        *sls_sources,
        sources=src["sls_sources"],
    )
    # Add the newly resolved blocks to the temporary run
    await hub.idem.sls_source.init.update(temporary_run_name, gather_data)

    if hub.idem.RUNS[temporary_run_name]["errors"]:
        result["comment"] = [
            f"Error in gathering sls_sources for sls.run {name}"
        ] + hub.idem.RUNS[temporary_run_name]["errors"]
        result["result"] = False
        hub.idem.RUNS.pop(temporary_run_name, None)
        return result

    # loop through high data and append sls.run name to all the states to make them unique
    # if the sls.run state is run on same set of files with different parameters adding sls.run name to all included
    # states makes them unique in idem run.
    modified_high_data = {}
    high_data = hub.idem.RUNS[temporary_run_name]["high"]
    for resource_name, resource_state in high_data.items():
        # Add sls_sources path for nested sls.run so that this path can be used as
        # base path for resolving sls_sources in nested sls.run
        if "sls" in resource_state:
            resource_state["sls"].append(
                {
                    "sls_sources_path": _get_sls_file_source_path(
                        sls_source_base_path,
                        resource_state["__sls__"],
                        hub.OPT.idem.tree,
                    )
                }
            )
        # append sls.run state name to all states in high data
        modified_high_data[f"{name}.{resource_name}"] = resource_state

    hub.idem.RUNS[temporary_run_name]["high"] = modified_high_data

    # compile the high data to low data
    await hub.idem.state.compile(temporary_run_name)

    _format_requisites_in_low_data(hub, name, temporary_run_name)

    low_data = hub.idem.RUNS[temporary_run_name]["low"]

    # Iterate over the states passed to this sls.run and add the extra attributes __sls_run_idm, sls_run_id
    # to identify these states from other states that are running in main idem run.
    for chunk in low_data:
        # Add the low data gathered from sls_sources passed to this state to main run
        hub.idem.RUNS[main_run_name].get("add_low").append(chunk)

    result["new_state"] = {
        "Status": f"Added {len(low_data)} states to be run.",
        "is_sls_run": True,
    }
    hub.idem.RUNS.pop(temporary_run_name, None)
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """
    Always skip reconciliation for sls.run because is_pending will be called for individual states
    """
    return False


# Utility methods to extract params and sls_sources passed to sls.run and update requisites


async def _gather_params(
    hub, main_run_name, temporary_run_name, params, param_sources_path, kwargs
):
    """Gather params passed to main Idem run, params files passed to the sls.run
       and inline parameters passed to sls.run
       will combine the params from all the three sources. If there are common params, Inline params takes precedence
       followed by params files provided to this state.

    Args:
        main_run_name(string):
            Name of the Idem run

        params(Dict[str, List[str]]):
            params sources and params files to get params from.

        param_sources_path(list):
            List of file paths of provided params

        kwargs(Dict[str, Any]):
            Inline params provided to the sls.run
    """
    run_params = {}
    # params given during Idem main run.
    run_params.update(hub.idem.RUNS[main_run_name]["params"] or {})

    # parse params files provide to this state.
    if params:
        params = hub.idem.sls_source.param.get_refs(
            sources=param_sources_path, refs=params
        )
        resolved_params_ret = await _gather_params_from_included_files(
            hub, temporary_run_name, params
        )
        if not resolved_params_ret or "errors" in resolved_params_ret:
            return {"errors": resolved_params_ret["errors"]}
        # combining the params provided to this file and idem run params.
        # if there is overlapping of params, params provided to this file will take precedence over idem run params.
        if "params" in resolved_params_ret:
            run_params.update(resolved_params_ret["params"])

    # update the params with inline params if provided.
    if kwargs:
        run_params.update(kwargs)
    hub.idem.RUNS[temporary_run_name]["params"] = run_params
    return {"params": run_params}


async def _gather_params_from_included_files(hub, temporary_run_name, params):
    """Gather parameters from the params files.

    Args:
        temporary_run_name(string):
            Name of the Idem run

        params(Dict[str, List[str]]):
            params sources and params files to get params from.
    """

    param_sources = params["param_sources"]
    params = params["params"]
    gather_data = await hub.idem.resolve.init.gather(
        temporary_run_name, *params, sources=param_sources
    )
    if gather_data["errors"]:
        return {"errors": gather_data["errors"]}
    hub.idem.sls_source.param.process_params(temporary_run_name, gather_data)

    return {"params": hub.idem.RUNS[temporary_run_name]["params"]}


def _format_requisites_in_low_data(hub, name, run_name):
    """
    In this function we add sls.run state name to the requisites.
    we loop through the low data find the requisites and add sls.run state name to
    identify the exact state we are referring
    """
    low_data = hub.idem.RUNS[run_name]["low"]
    # loop through low data
    for state in low_data:
        # loop through resource attributes
        for attribute_key, attribute_val in state.items():
            # check if the requisite is present
            if attribute_key in REQUISITES_TO_UPDATE_FOR_SLS_RUN:
                for require_state in attribute_val:
                    updated_requisites = []
                    state_func = next(iter(require_state))
                    require_arg_value = require_state[state_func]

                    if isinstance(require_arg_value, list):
                        name_defs = require_arg_value
                    else:
                        name_defs = [{require_arg_value: []}]

                    for requisite in name_defs:
                        state_name = next(iter(requisite))
                        args = requisite[state_name]
                        # check if the referred requisite is part of this current run
                        # if chunk is part of this sls.run add sls.run name else do not add sls.run name
                        r_chunks = hub.idem.tools.get_chunks(
                            low_data, state_func, f"{name}.{state_name}"
                        )
                        if r_chunks:
                            updated_requisites.append({f"{name}.{state_name}": args})
                            # update the sls.run name in the argument binding references
                            if attribute_key == "arg_bind":
                                for arg in args:
                                    req_key_path = next(iter(arg))
                                    chunk_key_path = arg[req_key_path]
                                    arg_reference_def = (
                                        "${"
                                        + f"{state_func}:{state_name}:{req_key_path}"
                                        + "}"
                                    )
                                    updated_arg_reference_def = (
                                        "${"
                                        + f"{state_func}:{name}.{state_name}:{req_key_path}"
                                        + "}"
                                    )
                                    hub.idem.rules.arg_resolver.set_chunk_arg_value(
                                        state,
                                        arg_reference_def,
                                        chunk_key_path.split(":"),
                                        updated_arg_reference_def,
                                        None,
                                    )
                        else:
                            updated_requisites.append({f"{state_name}": args})

                    require_state[state_func] = updated_requisites
    hub.idem.RUNS[run_name]["low"] = low_data


def _get_sls_file_source_path(
    source_paths: str, sls_file_path: str, base_path: str = ""
):
    """
    This method returns the full path by combining source path and file path of parent sls.run,
    so that this path can be used as source path for nested sls.run states.
    """
    updated_paths = []
    for source_path in source_paths:
        source_path = "/".join(source_path.split("/")[0:-1])
        updated_paths.append(os.path.join(base_path, source_path, sls_file_path))
    return tuple(updated_paths)



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the multi stage routine to determine the current requisite sequence
    """
    for seq_plugin in sorted(hub.idem.req.seq._loaded.keys()):
        if seq_plugin == "init":
            continue
        seq = hub.idem.req.seq[seq_plugin].run(seq, low, running, options)
    return seq



from typing import Any
from typing import Dict
from typing import List
from typing import Set

KEYWORDS = ["extend", "__extend__"]


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
    is_params: bool,
) -> List[str]:
    """
    Resolve the extend statement

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    :param is_params: True for parameter sources
    """
    if "extend" not in state:
        return []
    ext = state.pop("extend", {})
    if not isinstance(ext, Dict):
        hub.idem.RUNS[name]["errors"].append(
            f'Extension value in SLS "{sls_ref}" is not a dictionary'
        )
        return
    for id_ in ext:
        if not isinstance(ext[id_], Dict):
            hub.idem.RUNS[name]["errors"].append(
                f'Extension ID "{id_}" in SLS "{sls_ref}" is not a dictionary'
            )
            continue
        if "__sls__" not in ext[id_]:
            ext[id_]["__sls__"] = sls_ref
        # if '__env__' not in ext[id_]:
        #    ext[id_]['__env__'] = saltenv
        for key in list(ext[id_]):
            if key.startswith("_"):
                continue
            if not isinstance(ext[id_][key], list):
                continue
            if "." in key:
                comps = key.split(".")
                ext[id_][comps[0]] = ext[id_].pop(key)
                ext[id_][comps[0]].append(comps[1])
    if ext:
        state.setdefault("__extend__", []).append(ext)

    return []



import copy
import time


def order_chunks(hub, chunks):
    """
    Sort the chunk list verifying that the chunks follow the order
    specified in the order options.
    """
    cap = 1
    for chunk in chunks:
        if "order" in chunk:
            if not isinstance(chunk["order"], int):
                continue

            chunk_order = chunk["order"]
            if chunk_order > cap - 1 and chunk_order > 0:
                cap = chunk_order + 100
    for chunk in chunks:
        if "order" not in chunk:
            chunk["order"] = cap
            cap += 1
            continue

        if not isinstance(chunk["order"], (int, float)):
            if chunk["order"] == "last":
                chunk["order"] = cap + 1000000
            elif chunk["order"] == "first":
                chunk["order"] = 0
            else:
                chunk["order"] = cap
                cap += 1
        if "name_order" in chunk:
            chunk["order"] = chunk["order"] + chunk.pop("name_order") / 10000.0
        if chunk["order"] < 0:
            chunk["order"] = cap + 1000000 + chunk["order"]
    chunks.sort(
        key=lambda chunk: (chunk["order"], "{0[state]}{0[name]}{0[fun]}".format(chunk))
    )
    return chunks


def compile(hub, high, add_low):
    """
    "Compile" the high data as it is retrieved from the CLI or YAML into
    the individual state executor structures
    """
    chunks = []
    for name, body in high.items():
        if name.startswith("__"):
            continue
        for state, run in body.items():
            funcs = set()
            names = []
            if state.startswith("__"):
                continue
            chunk = {}
            chunk["state"] = state
            chunk["name"] = name
            if "__sls__" in body:
                chunk["__sls__"] = body["__sls__"]
            if "__env__" in body:
                chunk["__env__"] = body["__env__"]
            chunk["__id__"] = name
            for arg in run:
                if isinstance(arg, str):
                    funcs.add(arg)
                    continue
                if isinstance(arg, dict):
                    for key, val in arg.items():
                        if key == "names":
                            for _name in val:
                                if _name not in names:
                                    names.append(_name)
                        elif key == "state":
                            # Don't pass down a state override
                            continue
                        elif key == "name" and not isinstance(val, str):
                            # Invalid name, fall back to ID
                            chunk[key] = name
                        elif key == "name_prefix" and chunk["name"] == name:
                            # the flow will be as follows:
                            # 1. if `name` is passed explicitly in sls, we should consider it.
                            # 2. if `name` is not passed but the `name_prefix` is passed in sls, then we should create a
                            # unique name beginning with the specified prefix.
                            # 3. if none of them is mentioned, it falls to default behaviour (value will be same as ID).

                            chunk["name"] = val + str(time.time())
                            chunk[key] = val
                        else:
                            chunk[key] = val
            if names:
                name_order = 1
                for entry in names:
                    live = copy.deepcopy(chunk)
                    if isinstance(entry, dict):
                        low_name = next(entry.keys())
                        live["name"] = low_name
                        list(map(live.update, entry[low_name]))
                    else:
                        live["name"] = entry
                    live["name_order"] = name_order
                    name_order += 1
                    for fun in funcs:
                        live["fun"] = fun
                        chunks.append(live)
            else:
                live = copy.deepcopy(chunk)
                for fun in funcs:
                    live["fun"] = fun
                    chunks.append(live)
    chunks.extend(add_low)
    chunks = hub.idem.ccomps.low.order_chunks(chunks)
    return chunks



# Implementing this contract will catch any exception from any exec module function
import asyncio


def _run_catching_errors(ctx):
    try:
        return ctx.func(*ctx.args, **ctx.kwargs)
    except Exception as e:
        return _format_error(e)


async def _arun_catching_errors(hub, ctx):
    try:
        ret = ctx.func(*ctx.args, **ctx.kwargs)
        return await hub.pop.loop.unwrap(ret)
    except Exception as e:
        return _format_error(e)


def _format_error(e: Exception):
    return dict(comment=[f"{e.__class__.__name__}: {e}"], result=False, ret=None)


def call(hub, ctx):
    if asyncio.iscoroutinefunction(ctx.func):
        return _arun_catching_errors(hub, ctx)
    else:
        return _run_catching_errors(ctx)



import random


def get(hub, min_value: int = None, max_value: int = None, **kwargs):
    # return wait time in seconds which is a random number
    # between min and max (inclusive) values
    #
    # Define in state:
    # __reconcile_wait__ = {"random": {"min_value": 1, "max_value": 10}}

    if min_value is None or max_value is None or min_value >= max_value:
        raise ValueError(
            "Expecting min and max values for the randomly generated wait time in seconds. "
            "For example: min_value=1, max_value=10."
        )

    val = random.randint(min_value, max_value)
    hub.log.debug(f"Random wait time (seconds) between {min_value}-{max_value}: {val}")
    return val



from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Parse argument binding reference and update the chunk argument with the requisite value.
    For example:
    - arg_bind:
        - cloud:
            - referenced_state:
                - referenced_state_property: this_state_property
    """
    if condition != "arg_bind":
        return {"errors": [f'"{condition}" is not a supported arg resolver.']}

    if not isinstance(reqret.get("args", {}), list):
        return {"errors": [f'"{condition}" is not in a supported format.']}

    # TODO: Add debug logging
    # Iterate over args of the "arg_bind" requisite, each arg is an argument binding
    for req_def in reqret.get("args", []):
        # Validate that requisite definition is a key/value pair,
        # where the key is the referenced state argument path
        # and the value is the current state argument path
        if not isinstance(req_def, dict):
            return {"errors": [f'"{req_def}" is not in a supported format.']}
        req_key_path = next(iter(req_def))
        chunk_key_path = req_def[req_key_path]
        resource_state = reqret.get("state")
        resource_name = reqret.get("name")

        if reqret["ret"].get("new_state", None) is None:
            return {
                "errors": [
                    f'"{resource_state}:{resource_name}" state does not have "new_state" in the state returns.'
                ]
            }

        # Construct state argument reference definition based on arg_bind requisite
        arg_reference_def = (
            "${" + f"{resource_state}:{resource_name}:{req_key_path}" + "}"
        )

        try:
            # First, find the value of the referenced state argument, based on the path
            # Iterate over referenced state argument key chain and find the value in the state "new_state"
            req_arg_value = hub.tool.idem.arg_bind_utils.parse_dict_and_list(
                resource_state, reqret["ret"]["new_state"], req_key_path.split(":"), ctx
            )

            # Second, set current state argument to the referenced state argument value
            hub.log.debug(
                f"Replacing references to `{arg_reference_def}` with value `{req_arg_value}`"
            )

            hub.idem.rules.arg_resolver.set_chunk_arg_value(
                chunk,
                arg_reference_def,
                chunk_key_path.split(":"),
                req_arg_value,
                None,
            )
        except IndexError as ex:
            return {"errors": [f"{ex}"]}
        except ValueError as ex:
            return {"errors": [f"{ex}"]}

    return {}


def set_chunk_arg_value(
    hub, chunk, arg_reference_def, arg_key_chain, arg_value, chunk_indexes
):
    """
    Recursively iterate over arg_keys and update the chunk desired key with the referenced value
    """
    arg_key = arg_key_chain.pop(0)
    arg_key, next_chunk_indexes = hub.tool.idem.arg_bind_utils.parse_index(arg_key)
    # Unescape dictionary references in the key definition.
    arg_key = arg_key.replace("[\\", "[")

    if len(arg_key_chain) == 0:
        indexed_chunk = hub.tool.idem.arg_bind_utils.get_chunk_with_index(
            chunk, chunk_indexes, arg_key
        )

        if next_chunk_indexes:
            # arg_key is set to a (nested) collection and arg_value is added to it , ex: arg_key[0][1][2]
            _set_chunk_with_index(
                indexed_chunk[arg_key], next_chunk_indexes, arg_reference_def, arg_value
            )
        else:
            if arg_key not in indexed_chunk:
                indexed_chunk[arg_key] = ""
            # arg_key is set to arg_value
            indexed_chunk[arg_key] = _replace_arg_reference_with_arg_value(
                indexed_chunk[arg_key], arg_reference_def, arg_value
            )

    else:
        chunk = hub.tool.idem.arg_bind_utils.get_chunk_with_index(
            chunk, chunk_indexes, arg_key
        )
        if arg_key not in chunk:
            chunk[arg_key] = {}

        hub.idem.rules.arg_resolver.set_chunk_arg_value(
            chunk[arg_key],
            arg_reference_def,
            arg_key_chain,
            arg_value,
            next_chunk_indexes,
        )


def _set_chunk_with_index(chunk, chunk_indexes, arg_reference_def, arg_value):
    index = chunk_indexes.pop(0)
    if not index.isdigit():
        raise ValueError(
            f'Cannot set argument value for index "{index}". The "{index}" is not supported'
        )
    index = int(index)
    if not isinstance(chunk, list) or len(chunk) < index + 1:
        raise ValueError(
            f'Cannot set argument value for index "{index}", '
            f'because argument key is not a list or it does not include element with index "{index}".'
        )

    if len(chunk_indexes) == 0:
        chunk[index] = _replace_arg_reference_with_arg_value(
            chunk[index], arg_reference_def, arg_value
        )
    else:
        _set_chunk_with_index(chunk[index], chunk_indexes, arg_reference_def, arg_value)


def _replace_arg_reference_with_arg_value(
    chunk_arg_value, arg_reference_def, arg_value
):
    """
    Find all occurrences of arg_reference_def in chunk_arg_value and replace them with arg_value.
    Return chunk_arg_value with all references to arg_reference_def resolved, or arg_value if no
    references found.
    """
    if (
        chunk_arg_value
        and isinstance(chunk_arg_value, str)
        and isinstance(arg_value, str)
        and arg_reference_def in chunk_arg_value
    ):
        return str.replace(chunk_arg_value, arg_reference_def, arg_value)
    else:
        return arg_value



import subprocess
import sys
from typing import Dict
from typing import Tuple

try:
    import pip
    from pip._internal.operations.freeze import freeze as freeze_

    HAS_PIP = True
except ImportError:
    HAS_PIP = False


def freeze(hub) -> Dict[str, str]:
    """
    Return the output of pip freeze
    """
    if HAS_PIP:
        # Get the freeze information from pip internals
        pip_freeze = [line for line in freeze_()]
    else:
        # Shell out to pip
        ret = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
        pip_freeze = ret.decode("utf-8").strip().split("\n")

    result = {}
    for line in pip_freeze:
        version_ = ""
        package = ""
        editable = "-e " in line

        if "#egg=" in line:
            v, package = line.split("#egg=")
        elif "==" in line:
            package, v = line.split("==")

        result[package] = {"version": version_, "editable": editable}
    return result


def version(hub) -> Tuple[int, int, int]:
    """
    Return the version of pip
    """
    major = minor = point = 0
    try:
        split = pip.__version__.split(".", maxsplit=2)
        major = int(split[0])
        minor = int(split[1])
        point = int(split[2])
    except:
        ...

    return major, minor, point



from idem.exec.init import ExecReturn

try:
    import colorama

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def display(hub, data):
    """
    Display the data from an idem run
    """
    if not isinstance(data, ExecReturn):
        return hub.output.nested.display(data)

    if not data.result:
        return colorama.Fore.RED + str(data.comment)
    else:
        return hub.output.nested.display(data.ret)



import copy
from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Check to see if the reqret state will make changes in the future
    """
    # condition = "changes"
    # First, prepare the ctx for the reqret chunk
    # Execute the reqret chunk in test mode
    # if the condition returns True then no errors are presented
    ret = {"errors": []}
    ctx = copy.deepcopy(ctx)
    ctx["test"] = True
    r_chunk = copy.deepcopy(reqret["chunk"])
    func = hub.idem.rules.init.get_func(name, r_chunk)
    r_chunk["ctx"] = ctx
    call_kwargs = hub.idem.tools.format_call(
        fun=func,
        data=r_chunk,
        expected_extra_kws=hub.idem.rules.init.STATE_INTERNAL_KEYWORDS,
    )
    sret = func(**call_kwargs)
    cond = sret.get(condition, False)
    if not cond:
        ret["errors"].append(
            f"No {condition} found in execution of {reqret['state']}.{reqret['name']}"
        )
    return ret



import re
from typing import Any
from typing import Dict
from typing import List
from typing import Set

KEYWORDS = ["include"]


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
    is_params: bool,
) -> List[str]:
    """
    Parse through the includes and download not-yet-resolved includes

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    :param is_params: True for parameter sources
    """
    # Maintain unresolved as a list to keep the sls file order. This order is used during param override.
    unresolved = []
    include = state.pop("include", [])

    if not isinstance(include, List):
        hub.idem.RUNS[name]["errors"].append(
            f"Include Declaration in SLS {sls_ref} is not formed as a list but as a {type(include)}"
        )
        return unresolved

    # Handle relative includes
    for inc_sls in include:
        # If the inc_sls starts with a dot then it is a relative include
        if inc_sls.startswith("."):
            match = re.match(r"^(\.+)(.*)$", inc_sls)

            # "levels" indicates the number of parents that need to be considered relative to the resolved source
            levels, inc_ref = match.groups()

            # The number of "."s indicates how many parents of the sls_ref we need to remove
            level_count = len(levels)

            # Transform the relative ref into an absolute ref
            p_comps = sls_ref.split(".")
            if level_count > len(p_comps):
                hub.idem.RUNS[name]["errors"].append(
                    f'Attempted relative include of "{inc_sls}" within SLS {sls_ref} goes beyond top level package'
                )
                continue
            inc_sls = ".".join(p_comps[:-level_count] + [inc_ref])

        # Make sure unresolved contains a unique set of sls files' reference
        if inc_sls not in resolved and inc_sls not in unresolved:
            unresolved.append(inc_sls)

    return unresolved



import asyncio


# Implementing this contract will catch any exception from any state


def _format_error(ctx, e):
    return {
        "changes": {},
        "comment": [f"{e.__class__.__name__}: {e}"],
        "name": ctx.kwargs.get("name"),
        "result": False,
    }


async def _catch_all(hub, ctx, ret):
    try:
        return await hub.pop.loop.unwrap(ret)
    except Exception as e:
        return _format_error(ctx, e)


def call(hub, ctx):
    try:
        ret = ctx.func(*ctx.args, **ctx.kwargs)
        if asyncio.iscoroutine(ret):
            return _catch_all(hub, ctx, ret)
        else:
            return ret
    except Exception as e:
        return _format_error(ctx, e)



# An exec module that can run states


async def run(
    hub,
    path: str,
    name: str = None,
    acct_profile: str = None,
    test: bool = False,
    *args,
    **kwargs,
):
    """
    Trigger a state manually from the CLI.
    This skips ESM, Reconciliation, and some parts of State contracts.
    This is useful in scenarios when you want to run a state directly without any file inputs.

    Generic example:

    .. code-block:: bash

        $ idem exec state.run <state.ref> kwarg1=value1 test=True acct_profile=default

    Specific example:

    .. code-block:: bash

        $ idem exec state.run time.sleep duration=1 test=True --output=json
    """
    # Construct the ctx that a state uses
    ctx = await hub.idem.acct.ctx(
        f"states.{path}",
        profile=acct_profile or hub.OPT.idem.get("acct_profile"),
        acct_key=hub.OPT.acct.get("acct_key"),
        acct_file=hub.OPT.acct.get("acct_file"),
        hard_fail=True,
        validate=True,
    )
    # Override ctx.test with the value from the parameters
    ctx.test = test

    # Call the state with the given parameters and await the result
    ret = await hub.pop.loop.unwrap(hub.states[path](ctx, name=name, *args, **kwargs))

    return {
        "result": ret["result"],
        "comment": ret["comment"],
        "ret": ret["new_state"],
        "test": test,
    }



from idem.reconcile.init import DEFAULT_MAX_PENDING_RERUNS


# Reconciliation loop stops after MAX_RERUNS_WO_CHANGE reruns without change for failed states.
# This is to make sure we do not retry forever on failures that cannot be fixed by
# reconciliation.
MAX_RERUNS_WO_CHANGE = 3


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """
    Default implementation of pending plugin
    Returns 'True' when the state is still 'pending' and reconciliation is required.

    If the state implements is_pending() call the state's implementation,
    otherwise state is pending until 'result' is 'True' and there are no 'changes'
    for normal flows, state is pending until 'result' is 'True' for resource recreation flow.
    Return false if the last consecutive runs produced the same result ('False') and 'changes',
    to stop reconciliation.

    :param hub: The hub
    :param ret: (dict) Returned structure of a run
    :param state: (Text, Optional) The name of the state
    :param pending_kwargs: (dict, Optional) May include 'reruns_wo_change_count' 'reruns_count' 'max_pending_reruns'
    :return: bool
    """
    # When the state implements is_pending, we should let it determine when
    # to stop the reconciliation re-runs. However we should stop after time to make
    # sure we do not reconcile forever.
    max_reruns = (
        pending_kwargs.get("max_pending_reruns", DEFAULT_MAX_PENDING_RERUNS)
        if pending_kwargs
        else DEFAULT_MAX_PENDING_RERUNS
    )

    if (
        state is not None
        and hub.states[state] is not None
        and callable(getattr(hub.states[state], "is_pending", None))
    ):
        # For safety we should limit re-runs with a very large number
        if pending_kwargs and pending_kwargs.get("reruns_count", 0) >= max_reruns:
            return False

        return hub.states[state].is_pending(ret=ret, state=state, **pending_kwargs)

    # We should not execute errors more than three times w/o changes in result
    if (
        pending_kwargs
        and pending_kwargs.get("reruns_wo_change_count", 0) >= MAX_RERUNS_WO_CHANGE
    ):
        return False

    # for resource recreation flow, we should consider only whether the result is True or not.
    if ret.get("recreation_flow", False):
        return not ret["result"] is True

    return not ret["result"] is True or bool(ret["changes"])



def reconcile(hub, high):
    """
    Take the extend statement and reconcile it back into the highdata
    """
    errors = []
    if "__extend__" not in high:
        return high, errors
    ext = high.pop("__extend__")
    for ext_chunk in ext:
        for id_, body in ext_chunk.items():
            if id_ not in high:
                state_type = next(x for x in body if not x.startswith("__"))
                # Check for a matching 'name' override in high data
                ids = hub.idem.tools.find_id(id_, state_type, high)
                if len(ids) != 1:
                    errors.append(
                        "Cannot extend ID '{0}' in '{1}:{2}'. It is not "
                        "part of the high state.\n"
                        "This is likely due to a missing include statement "
                        "or an incorrectly typed ID.\nEnsure that a "
                        "state with an ID of '{0}' is available\nin "
                        "environment '{1}' and to SLS '{2}'".format(
                            id_,
                            body.get("__env__", "base"),
                            body.get("__sls__", "base"),
                        )
                    )
                    continue
                else:
                    id_ = ids[0][0]

            for state, run in body.items():
                if state.startswith("__"):
                    continue
                if state not in high[id_]:
                    high[id_][state] = run
                    continue
                for arg in run:
                    update = False
                    for hind in range(len(high[id_][state])):
                        if isinstance(arg, str) and isinstance(
                            high[id_][state][hind], str
                        ):
                            # replacing the function, replace the index
                            high[id_][state].pop(hind)
                            high[id_][state].insert(hind, arg)
                            update = True
                            continue
                        if isinstance(arg, dict) and isinstance(
                            high[id_][state][hind], dict
                        ):
                            # It is an option, make sure the options match
                            argfirst = next(iter(arg))
                            if argfirst == next(iter(high[id_][state][hind])):
                                # If argfirst is a requisite then we must merge
                                # our requisite with that of the target state
                                if (
                                    argfirst
                                    in hub.idem.rules.init.STATE_REQUISITE_KEYWORDS
                                ):
                                    high[id_][state][hind][argfirst].extend(
                                        arg[argfirst]
                                    )
                                # otherwise, its not a requisite and we are just extending (replacing)
                                else:
                                    high[id_][state][hind] = arg
                                update = True
                            if (
                                argfirst == "name"
                                and next(iter(high[id_][state][hind])) == "names"
                            ):
                                # If names are overwritten by name use the name
                                high[id_][state][hind] = arg
                    if not update:
                        high[id_][state].append(arg)
    return high, errors



import pathlib
import tempfile

__contracts__ = ["soft_fail"]

SKIP_ESM = True


async def block(hub, ctx, name: str, sls: str = None, render: str = None):
    """
    Render a block of sls and add it to the state runs

    .. code-block:: sls

        delay:
          time.sleep:
            - duration: 2

        delayed_render_block:
          render.block:
            - block: >
               # Use "raw" statements so jinja doesn't process this block in the first pass through the renderer
               {% raw %}
               # new state to be processed

               new_state:
                 exec.run:
                   - path: test.ping

               {% endraw %}
            # Only process these new states after other requirements have been met
            - require:
               - time: delay

    """
    result = dict(comment=[], new_state={}, old_state=None, name=name, result=True)
    # Create a temporary run name to use while compiling the new sls block
    temporary_run = f"delayed-render-{name}"
    main_run = ctx.run_name

    # Write the new sls block to a temporary to a temporary file
    # So that it can be processed the same way as input sls
    with tempfile.NamedTemporaryFile(suffix=".sls", delete=True) as fh:
        fh.write(sls.encode())
        fh.flush()
        path = pathlib.Path(fh.name)
        # Pass through other sources from the main run to resolve includes in the new sls with existing sources
        sources = [f"file://{path.parent.absolute()}"] + hub.idem.RUNS[main_run][
            "sls_sources"
        ]

        # Create a temporary run to process the new sls block
        await hub.idem.state.create(
            name=temporary_run,
            sls_sources=sources,
            # Allow a different render pipe to be used for the new render block
            # Default to the renderer of the main run
            render=render or hub.idem.RUNS[main_run].get("render"),
            # Copy state.apply parameters from the main run
            **{
                k: hub.idem.RUNS[main_run].get(k)
                for k in (
                    "runtime",
                    "subs",
                    "cache_dir",
                    "test",
                    "acct_file",
                    "acct_key",
                    "acct_profile",
                    "acct_blob",
                    "managed_state",
                    "param_sources",
                    "acct_data",
                    "invert_state",
                )
            },
        )
        try:
            # Resolve the new sls block with the main run
            gather_data = await hub.idem.resolve.init.gather(
                main_run, f"{path.stem}", sources=sources
            )
            # Add the newly resolved blocks to the temporary run
            await hub.idem.sls_source.init.update(temporary_run, gather_data)

            # Report any errors from compiling the new render block
            if hub.idem.RUNS[temporary_run]["errors"]:
                result["result"] = False
                result["comment"].extend(hub.idem.RUNS[temporary_run]["errors"])
            else:
                # Iterate over the states added to the temporary run
                for new_name, new_state in hub.idem.RUNS[temporary_run]["high"].items():
                    # If a state name from the new render block has the same name as a state in the main run, then fail
                    if new_name in hub.idem.RUNS[main_run]["high"]:
                        result["result"] = False
                        result["comment"].append(
                            f"State named '{new_state}' already exists in the run"
                        )
                        # Don't add the duplicate state to the main run
                        continue
                    else:
                        # Report the state that was added to the run in this state's output
                        result["new_state"][new_name] = new_state
                        # Add the state from the new render block to the main run
                        hub.idem.RUNS[main_run]["high"][new_name] = new_state
        finally:
            # Clear the temporary run we created
            hub.idem.RUNS.pop(temporary_run, None)

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """
    Always skip reconciliation for this plugin
    """
    return False



import asyncio


def post(hub, ctx):
    """
    Convert the dict return to an immutable namespace addressable format
    """
    ref = f"{ctx.ref}.{ctx.func.__name__}"
    kwargs = ctx.get_arguments()
    if asyncio.iscoroutine(ctx.ret):
        return hub.tool.idem.exec.format_async_return(
            ret=ctx.ret, ref=ref, kwargs=kwargs
        )
    else:
        return hub.tool.idem.exec.format_return(ret=ctx.ret, ref=ref, kwargs=kwargs)



from typing import Any
from typing import Dict


def define(hub) -> Dict[str, Any]:
    """
    Gather the requisite restrictions and populate the requisite behavior map
    """
    rmap = {}
    for mod in hub.idem.req._loaded:
        if mod == "init":
            continue
        rmap[mod] = hub.idem.req[mod].define()
    return rmap



from typing import Any
from typing import Dict


async def loop(
    hub,
    pending_plugin: str,
    max_pending_reruns: int,
    name: str,
    apply_kwargs: Dict[str, Any],
):
    # This is the default reconciler
    # Reconciliation loop is skipped for backward compatibility
    return {
        "re_runs_count": 0,
        "require_re_run": False,
    }



def sig_get(hub, **kwargs):
    """
    Validate the signature of the reconcile.wait get function
    """
    ...


def post_get(hub, ctx):
    """
    Conform the return value of every reconcile.wait get()
    """
    assert isinstance(ctx.ret, int)
    return ctx.ret



from typing import Any
from typing import Dict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Check the state containing the target func and call the aggregate
    function if present
    """
    func = hub.idem.rules.init.get_func(name, chunk, "mod_aggregate")
    if func:
        chunk = func(name, chunk)
        chunk = await hub.pop.loop.unwrap(chunk)
    return chunk



from typing import Any
from typing import Dict


def sig_apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    ret = {}
    for tag in data:
        state_ret = data[tag]
        comps = tag.split("_|-")
        state = comps[0]
        id_ = comps[1]
        fun = comps[3]
        result = state_ret.get("result")
        comment = state_ret.get("comment")
        changes = state_ret.get("changes", {})
        old_state = state_ret.get("old_state", {})
        new_state = state_ret.get("new_state", {})

        ret[tag] = data[tag]
    return ret
    """



import copy
from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    If changes are made then run the configured post command
    """
    ret = {}
    if reqret["ret"]["changes"]:
        id_ = f"{chunk['__id__']}_listen"
        post_chunk = copy.deepcopy(chunk)
        post_chunk["fun"] = condition
        post_chunk["__id__"] = id_
        post_chunk["name"] = id_
        post_chunk["order"] = -1
        hub.idem.RUNS[name]["post_low"].append(post_chunk)
    return ret



from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Check to see if there are changes
    """
    if isinstance(condition, bool):
        if bool(reqret["ret"]["changes"]) is condition:
            return {}
    # TODO: Add the ability to make more granular changes condition definitions
    elif reqret["ret"]["changes"] == condition:
        return {}
    return {
        "errors": [
            f'Changes from {reqret["r_tag"]} is "{bool(reqret["ret"]["changes"])}", not "{condition}"'
        ]
    }



# The order of the sequence that needs to be implemented:
# Start with a single sls file, just like you started with salt
# Stub out the routines around gathering the initial sls file
# Just use a yaml renderer and get it to where we can manage some basic
# includes to drive to highdata
# Then we can start to fill out renderers while at the same time
# deepening the compiler
import sys


def __init__(hub):
    # The modules to load config from
    hub.idem.CONFIG_LOAD = ["idem", "acct", "rend", "evbus", "pop_tree", "pop_loop"]
    hub.idem._omit_class = False
    hub.pop.sub.load_subdirs(hub.idem, recurse=True)
    hub.idem.RUNS = {}
    hub.pop.sub.add(dyne_name="log")
    hub.pop.sub.add(dyne_name="acct")
    hub.pop.sub.add(dyne_name="rend")
    hub.pop.sub.add(dyne_name="group")
    hub.pop.sub.add(dyne_name="output")
    hub.pop.sub.add(dyne_name="evbus")
    hub.pop.sub.add(dyne_name="reconcile")
    hub.pop.sub.add(dyne_name="source")
    hub.pop.sub.load_subdirs(hub.reconcile, recurse=True)
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    hub.pop.sub.add(dyne_name="esm")
    hub.pop.sub.add(dyne_name="exec")
    hub.pop.sub.load_subdirs(hub.exec, recurse=True)
    hub.pop.sub.add(dyne_name="states")
    hub.pop.sub.load_subdirs(hub.states, recurse=True)
    hub.pop.sub.add(dyne_name="tree", omit_class=False)
    hub.idem.RMAP = hub.idem.req.init.define()
    hub.idem.RUN_NAME = "cli"


def cli(hub):
    """
    Execute a single idem run from the cli
    """
    hub.pop.config.load(hub.idem.CONFIG_LOAD, cli="idem")

    # Check if modules are installed in editable mode with pip>=22
    try:
        pip_version = hub.tool.pip.version()
        if pip_version >= (22, 0, 0):
            freeze = hub.tool.pip.freeze()
            for module, metadata in freeze.items():
                if metadata["editable"] is True:
                    hub.log.warning(
                        f"'{module}' is installed in editable mode with pip>=22"
                    )
                    hub.log.warning(
                        f"install pip==21 or some modules may not load properly"
                    )
    except Exception as e:
        hub.log.error(f"Error checking pip freeze: {e.__class__.__name__}: {e}")

    # Initialize the async loop for the entire project
    hub.pop.loop.create()
    # Start the main program within the async loop
    retcode = hub.pop.Loop.run_until_complete(hub.idem.init.cli_apply())
    sys.exit(retcode)


async def cli_apply(hub) -> int:
    """
    Run the CLI routine in a loop
    """
    if hub.SUBPARSER in ("encrypt", "decrypt"):
        # Break early for acct commands
        return await hub.acct.init.cli_apply()
    elif hub.SUBPARSER == "doc":
        return await hub.idem.cli.doc()

    # Initialize the broker queue for evbus
    await hub.evbus.broker.init()

    # Specify the serializing plugin for evbus
    hub.evbus.SERIAL_PLUGIN = hub.OPT.evbus.serial_plugin
    # Use the run name as a routing key for exec modules
    hub.idem.RUN_NAME = hub.OPT.idem.run_name
    # Keep the local ESM cache file if explicitly configured to do so
    hub.idem.managed.KEEP_CACHE_FILE = hub.OPT.idem.esm_keep_cache

    # Collect ingress profiles from acct
    ingress_profiles = await hub.evbus.acct.profiles(
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
    )

    # Start the listener in it's own task
    listener = hub.pop.Loop.create_task(hub.evbus.init.start(ingress_profiles))
    try:
        await hub.evbus.init.join()

        if hub.SUBPARSER == "state":
            return await hub.idem.cli.sls()
        elif hub.SUBPARSER == "exec":
            return await hub.idem.cli.exec()
        elif hub.SUBPARSER == "describe":
            return await hub.idem.cli.desc()
        elif hub.SUBPARSER == "validate":
            return await hub.idem.cli.validate()
        elif hub.SUBPARSER == "refresh":
            return await hub.idem.cli.refresh()
        elif hub.SUBPARSER == "restore":
            return await hub.idem.cli.restore()
        else:
            print(hub.args.parser.help())
            return 2
    finally:
        await hub.evbus.init.stop()
        await listener



from typing import Any
from typing import Dict
from typing import List

DEFAULT_RESOLVER = "all"


def resolve(hub, rdats: Dict[str, List[Dict[str, Any]]]) -> List[str]:
    """
    Given the dict of requisite data sets, determine if the defined
    requisites passed. This allows for requisites to be defined
    as requiring, all, some, one, or none of the required states to
    pass
    """
    errors = []
    for req, requisite_list in rdats.items():
        requisite_map = hub.idem.RMAP[req]
        resolver = requisite_map.get("resolver", DEFAULT_RESOLVER)
        new_errors = hub.idem.resolver[resolver].check(requisite_list)
        errors.extend(new_errors)
    return errors


def check(hub, rlist: List[Dict[str, Any]]) -> List[str]:
    return []



from typing import Any
from typing import Dict
from typing import Iterable

import aiofiles
import dict_tools.data

__func_alias__ = {"ctx_": "ctx"}


async def ctx_(
    hub,
    path: str,
    profile: str,
    acct_file: str = None,
    acct_key: str = None,
    acct_blob: bytes = None,
    acct_data: Dict = None,
    hard_fail: bool = False,
    validate: bool = False,
):
    """
    :param hub:
    :param path: The reference on the hub to the function that needs authentication
    :param profile: The profile to use for this operation
    :param acct_file:
    :param acct_key:
    :param acct_blob:
    :param acct_data:
    :param hard_fail: Raise an error if there are any issues while collecting a profile
    :param validate: Raise an error if the function needs a ctx and the named profile does not exist
    :return:
    """
    ctx = dict_tools.data.NamespaceDict(test=False, acct=None, acct_details=None)

    parts = path.split(".")
    if parts[0] in ("exec", "states", "esm"):
        parts = parts[1:]
    elif parts[0] in ("idem",) and parts[1] in ("sls",):
        parts = parts[2:]

    if not parts:
        return ctx

    sname = parts[0]

    acct_paths = (
        f"exec.{sname}.ACCT",
        f"states.{sname}.ACCT",
        f"tool.{sname}.ACCT",
        f"esm.{sname}.ACCT",
        f"source.{sname}.ACCT",
        f"{sname}.ACCT",
    )

    if acct_data is None:
        acct_data = {}
    # Read encrypted profiles from an acct_file or acct_blob
    if acct_key:
        if acct_file:
            async with aiofiles.open(acct_file, "rb") as fh:
                acct_blob = await fh.read()
        if acct_blob:
            if acct_key:
                new_data = await hub.acct.init.unlock_blob(
                    acct_blob, acct_key=acct_key, hard_fail=hard_fail
                )
                acct_data.update(new_data)
    # read plaintext credentials from the acct_file
    elif acct_file:
        if "profiles" not in acct_data:
            acct_data["profiles"] = {}

        new_data = await hub.acct.init.profiles(acct_file, hard_fail=hard_fail)
        acct_data["profiles"].update(new_data)

    subs = set()
    for name in acct_paths:
        sub = getattr(hub, name, None)
        if not sub:
            continue
        if isinstance(sub, Iterable):
            subs.update(sub)

    acct = await hub.acct.init.gather(
        subs,
        profile,
        profiles=acct_data.get("profiles", {}),
        hard_fail=hard_fail,
    )

    if validate and profile and subs and not acct:
        # Validate that the specified acct_profile is specified is within the profiles
        profiles = set()
        for s in subs:
            profiles.update(acct_data.get("profiles", {}).get(s, {}).keys())

        if not profiles:
            raise ConnectionError(f"No profiles found for '{path}'")
        if profile not in profiles:
            raise ConnectionError(
                f"Could not find '{path}' profile '{profile}' in: {', '.join(profiles)}"
            )

    # The SafeNamespaceDict will not print its values, only keys
    ctx.acct = dict_tools.data.SafeNamespaceDict(acct)

    # Get metadata for the profile
    ctx.acct_details = await hub.acct.metadata.init.gather(
        profile=acct,
        providers=subs,
    )

    return ctx


async def data(hub, acct_key: str, acct_file: str, acct_blob: bytes) -> Dict[str, Any]:
    """
    Return a raw dictionary with all the profiles from an encrypted acct
    """
    acct_data = {}
    if acct_file:
        async with aiofiles.open(acct_file, "rb") as fh:
            acct_blob = await fh.read()
    if acct_blob:
        if acct_key:
            acct_data = await hub.acct.init.unlock_blob(acct_blob, acct_key=acct_key)
        else:
            if "profiles" not in acct_data:
                acct_data["profiles"] = {}
            acct_data["profiles"].update(await hub.acct.init.profiles(acct_file))
    return acct_data



import re

base_var_regex = r"[^\W0-9]\w*"  # Matches python identifier name
array_index = r"(\[(\d+)\])?"  # Matches array index

# Some example matches
#  ^^rg_name^^
#  ^^locations^^[0]
#  ^^lo_t10ns^^[19]
start_var = rf"\^\^({base_var_regex})\^\^({array_index})"

# Some example matches
#  ~~rg_name
#  ~~locations[0]
#  ~~lo_t10ns[19]
sub_var = rf"\.~~({base_var_regex})({array_index})"

# Compile the regular expressions
start_var_regex = re.compile(start_var)
sub_var_regex = re.compile(sub_var)

# Some match examples
#  ?? params.get('rg_name').get('your_rg', 'default') ?? ^^rg_name^^.~~your_rg ??
#  ?? params['locations'][0] ?? ^^locations^^[0] ??
#  ?? params['locations'][3].get('xstate') ?? ^^locations^^[3].~~xstate ??
extract_param_info_regex = re.compile(
    r"[?][?] .+?(?= [?][?]) [?][?] (.+?(?= [?][?])) [?][?]"
)


def _process_vars(params, input, pos):
    global start_var_regex

    # Here we are doing a regex match for strings like
    # ^^rg_name^^ or ^^locations^^[0]
    # That is a string starting and ending with double ^^,
    # with optional index. This gives us the parameter
    # name, and array index, if any.
    match = start_var_regex.search(input, pos)
    if match is None:
        return

    pos = _process_match(params, -1, match, input)
    _process_vars(params, input, pos)


def _fill_list(l, i):
    for i in range(len(l), i + 1):
        l.append(None)


# This function processes recursively any subparameters. For e.g.
# in the string   ^^rg_name^^.~~my_rg.~~abc   this function will
# process    .~~my_rg.~~abc
def _process_dot_vars(params, parent_index, param_name, index, input, pos):
    global sub_var_regex

    # Matchs a string like    ~~my_rg[3]    where array index
    # is optional.
    match = sub_var_regex.match(input, pos)

    # No match means we don't have sub parameter to process. For e.g.
    # in case of string being    ^^rg_name^^.~~my_rg.~~abc   we
    # have processed till abc and there is nothing to process further.
    if match is None:
        if isinstance(params, list):
            _fill_list(params, index)
            params[parent_index] = {param_name: ""}
        else:
            if param_name not in params:
                if index < 0:
                    params[param_name] = {}
                else:
                    params[param_name] = []
                    _fill_list(params[param_name], index)
            if index < 0:
                params[param_name] = ""
            else:
                params[param_name][index] = ""
        return pos

    if isinstance(params, list):
        if index < 0:
            o = {}
        else:
            o = []
            _fill_list(o, index)
        params[parent_index] = {param_name: o}
        return _process_match(o, index, match, input)
    else:
        if param_name not in params:
            if index < 0:
                params[param_name] = {}
            else:
                params[param_name] = []
                _fill_list(params[param_name], index)
        else:
            if index >= 0:
                _fill_list(params[param_name], index)

    return _process_match(params[param_name], index, match, input)


def _process_match(params, parent_index, match, input):
    groups = match.groups()
    param_name = groups[0]

    # A index is matched, which means we are dealing with array
    if len(groups) == 4 and groups[2] is not None:
        return _process_dot_vars(
            params, parent_index, param_name, int(groups[3]), input, match.end()
        )
    else:  # we are dealing with something that is not an array
        return _process_dot_vars(
            params, parent_index, param_name, -1, input, match.end()
        )


def _iterate_over_state(state, var):
    if isinstance(var, dict):
        for key, value in var.items():
            yield from _iterate_over_state(state, key)
            yield from _iterate_over_state(state, value)
    elif isinstance(var, list):
        for item in var:
            yield from _iterate_over_state(state, item)
    else:
        yield (state, var)


def _iterate_over(states):
    for state, value in states.items():
        sls = value["__sls__"]
        yield (f"{sls}.{state}", state)
        yield from _iterate_over_state(f"{sls}.{state}", value)


def _fill_list(l, i):
    for i in range(len(l), i + 1):
        l.append(None)


def stage(hub, name):
    all_params = {}
    # Here we iterate over all the strings in high data. Subsequently
    # in the _process_vars we process each such string, identifying
    # parameters (if any), in the context of a state.
    for state, value in _iterate_over(hub.idem.RUNS[name]["high"]):

        state = hub.tool.parameter.split_and_decode_base64(state)

        if not isinstance(value, str):
            continue

        if state not in all_params:
            all_params[state] = {}

        value = hub.tool.parameter.split_and_decode_base64(value)

        for param_info in extract_param_info_regex.findall(value):
            _process_vars(all_params[state], param_info, 0)

    hub.idem.RUNS[name]["parameters"] = {
        "GLOBAL": hub.idem.RUNS[name]["params"].params(),
        "ID_DECS": all_params,
    }



from typing import Any
from typing import Dict
from typing import List


def check(hub, rlist: List[Dict[str, Any]]) -> List[str]:
    """
    Chek the resolve list from the requisite systems and return errors from all
    errored requisites
    """
    errors = []
    for rdat in rlist:
        errors.extend(rdat.get("errors", []))
    return errors



from typing import Any
from typing import Dict


def apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Apply all group plugins specified in config
    """
    # Read groups from the config
    groups = hub.OPT.get("idem", {}).get("group")

    if not groups:
        return data

    # Run each of the group plugins from config, separated by a pipe
    for group_plugin in groups.split("|"):
        if group_plugin == "init":
            continue
        if group_plugin not in hub.group:
            hub.log.error(f"No group plugin '{group_plugin}'")
            continue
        data = hub.group[group_plugin].apply(data)

    return data



def stage(hub, name: str):
    """
    Invert present absent states if required
    """

    if (
        "invert_state" not in hub.idem.RUNS[name]
        or not hub.idem.RUNS[name]["invert_state"]
    ):
        hub.log.debug("Skipping invert as not enabled")
        return

    low = hub.idem.ccomps.invert.apply(hub.idem.RUNS[name]["low"])
    hub.idem.RUNS[name]["low"] = low



from typing import Any
from typing import Dict


def apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Sort the data by run number
    """
    result = {}
    for tag in sorted(data, key=lambda k: data[k].get("__run_num", 0)):
        result[tag] = data[tag]

    return result



from typing import Any
from typing import Dict

from dict_tools.data import SafeNamespaceDict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Check the state containing the target func and call the mod_creds
    function if present. Therefore gathering the list of creds systems
    to use
    """
    run_name = chunk["ctx"]["run_name"]

    # Get the acct_profile that is being used from the chunk and fallback to the acct_profile in the RUN
    profile = chunk.pop("acct_profile", hub.idem.RUNS[run_name]["acct_profile"])

    # Some idem tests or projects that call idem from pure python specify a sub other than "states" for state modules
    # Check for state modules being in alternate locations but default to "states"
    for sub in hub.idem.RUNS[run_name]["subs"]:
        try:
            # Get the keyword arguments of the state function that will be called
            function_parameters = hub[sub][chunk["state"]][
                chunk["fun"]
            ].signature.parameters
            break
        except AttributeError:
            ...
    else:
        function_parameters = {}

    # If the acct_profile was in the function parameters, then add back the acct_profile that is being used
    if "acct_profile" in function_parameters:
        chunk["acct_profile"] = profile

    # If acct_data was in the function parameters and in the chunk, then return the chunk without constructing ctx.acct
    # The function is going to decide what to do with the acct_data
    if "acct_data" in function_parameters and chunk.get("acct_data"):
        # Allow acct_data in the function parameters, but protect it from being exposed in events and logs
        chunk["acct_data"] = SafeNamespaceDict(chunk["acct_data"])
        return chunk

    # Retrieve the acct_data from the state chunk and fallback to the acct_data from the RUN
    acct_data = chunk.pop("acct_data", hub.idem.RUNS[run_name]["acct_data"])

    hub.log.debug(f"Loaded profile: {profile}")

    new_ctx = await hub.idem.acct.ctx(
        path=chunk["state"],
        profile=profile,
        acct_data=acct_data,
        hard_fail=True,
        validate=True,
    )
    chunk["ctx"]["acct"] = new_ctx["acct"]
    chunk["ctx"]["acct_details"] = new_ctx["acct_details"]

    return chunk



from typing import Any
from typing import Dict
from typing import List


def get_refs(hub, *, sources: List[str], refs: List[str]) -> Dict[str, List[str]]:
    """
    Determine where the param sls sources are

    :param hub:
    :param sources: sls-sources or params-sources
    :param refs: References to sls within the given sources
    """
    ret = {"param_sources": [], "params": []}

    if not refs:
        return ret

    for ref in refs:
        hub.idem.sls_source.init.process(ret["param_sources"], ret["params"], ref)

    ret["param_sources"].extend(sources)

    return ret


async def gather(hub, name: str, *sls):
    """
    Gather the named param references into the RUNS structure

    :param hub:
    :param name: The state run name
    :param sls: sls locations within sources
    """
    sources = hub.idem.RUNS[name]["param_sources"]
    gather_data = await hub.idem.resolve.init.gather(name, *sls, sources=sources)
    if gather_data["errors"]:
        hub.idem.RUNS[name]["errors"] = gather_data["errors"]
        return
    hub.idem.sls_source.param.process_params(name, gather_data)


def process_params(hub, name, gather_data: Dict[str, Any]):
    """Process the gathered params and set the params values in hub."""
    for sls_ref in gather_data["order"]:
        state = gather_data["state"][sls_ref]
        for id_ in hub.idem.resolve.init.iter(state):
            # In case of params_processing being true, such as when we are doing
            # parameter processing only: we will just read the key and value pairs and add
            # to hub.idem.RUNS[name]["params"] dict. Then we can access the parameter value
            # as {{ params.get('foo') }}
            hub.idem.RUNS[name]["params"][id_] = (
                state[id_] if state[id_] is not None else "__PYTHON_NONE__"
            )



from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Check to see if the result is True
    """
    if isinstance(condition, list):
        if reqret["ret"]["result"] in condition:
            return {}
    if reqret["ret"]["result"] is condition:
        return {}
    else:
        return {
            "errors": [
                f'Result of require {reqret["r_tag"]} is "{reqret["ret"]["result"]}", not "{condition}"'
            ]
        }



"""
This plugin is used to resolve transparent requisites and apply them to
the lowstate

.. code-block:: python

    TREQ = {
        "func_D": {
            "require": [
                "foo.bar.baz.func_A",
                "test.func_B",
            ],
            "soft_require": [
                "cheese.func_C",
            ],
        },
        "unique": ["present", "absent"],
    }
"""
import pop.loader


def gather(hub, subs, low):
    """
    Given the runtime name and the chunk in question, determine what function
    on the hub that can be run
    """
    ret = {}
    for chunk in low:
        s_ref = chunk["state"]
        if s_ref in ret:
            continue
        for sub in subs:
            test = f"{sub}.{s_ref}"
            try:
                mod = getattr(hub, test)
            except AttributeError:
                continue
            if not isinstance(mod, pop.loader.LoadedMod):
                continue
            if mod is None:
                continue
            if hasattr(mod, "TREQ"):
                ret.update({s_ref: mod.TREQ})
    return ret


def apply(hub, subs, low):
    """
    Look up the transparent requisites as defined in state modules and apply
    them to the respective low chunks
    """
    treq = hub.idem.ccomps.treq.gather(subs, low)
    for chunk in low:
        if not chunk["state"] in treq:
            continue
        if chunk["fun"] in treq[chunk["state"]]:
            rule = treq[chunk["state"]][chunk["fun"]]
            _add_rule_to_chunk(low, chunk, rule)
        if "unique" in treq[chunk["state"]]:
            functions = treq[chunk["state"]]["unique"]
            _add_unique_to_chunk(low, chunk, functions)
    return low


def _add_rule_to_chunk(low, chunk, rule):
    for rule_name, rule_refs in rule.items():
        for ref in rule_refs:
            for req_chunk in low:
                req_path = f'{req_chunk["state"]}.{req_chunk["fun"]}'
                if req_path == ref:
                    if rule_name not in chunk:
                        chunk[rule_name] = []
                    chunk[rule_name].append({req_chunk["state"]: req_chunk["__id__"]})


def _add_unique_to_chunk(low, chunk, functions):
    if chunk["fun"] in functions:
        chunk["unique"] = chunk["fun"]



import re
from typing import Dict
from typing import List


def parse_dict_and_list(
    hub, state_id: str, data: Dict or List, keys: List, ctx: Dict = None
):
    """
    Traverse a dict or list using a colon-delimited (or otherwise delimited,
    using the 'delimiter' param) target string. The target 'foo:bar:0' will
    return data['foo']['bar'][0] if this value exists, and will otherwise
    return the dict in the default argument.
    List resolution - The target 'foo:bar:[0]' or 'foo:bar[0]' will return data['foo']['bar'][0] if data
    like {'foo':{'bar':['baz']}}
    Dict resolution - The target 'foo:bar:0' will return data['foo']['bar'][0] if data like
    {'foo':{'bar':{'0':'baz'}}}
    """

    try:
        ref_path = ":".join(keys)
        ptr = data
        for key in keys:
            key_to_parse, indexes = hub.tool.idem.arg_bind_utils.parse_index(key)
            if isinstance(ptr, list):
                if key_to_parse:
                    ptr = [
                        parse_dict(hub, item, key_to_parse, indexes, ctx)
                        for item in ptr
                    ]
                else:
                    # If argument binding contains a nested collection, find the element in the collection, ex arg[0][1][2]
                    ptr = hub.tool.idem.arg_bind_utils.get_chunk_with_index(
                        ptr, indexes, ref_path
                    )
            else:
                ptr = parse_dict(hub, ptr, key_to_parse, indexes, ctx)
    except ValueError as ex:
        raise ValueError(f'"Failed to parse "{ref_path}" for state "{state_id}". {ex}')
    return ptr


def parse_dict(hub, data: Dict, key_to_parse: str, indexes: List, ctx: Dict = None):
    ptr = data
    if key_to_parse:
        if key_to_parse not in ptr:
            if ctx and ctx.get("test", False):
                return f"{key_to_parse}_value_known_after_applying"
            else:
                raise ValueError(
                    f'Key "{key_to_parse}" is not found as part of the state "new_state".'
                )
        ptr = ptr[key_to_parse]
        # If argument binding contains a nested collection, find the element in the collection, ex arg[0][1][2]
        ptr = hub.tool.idem.arg_bind_utils.get_chunk_with_index(
            ptr, indexes, key_to_parse
        )

    return ptr


async def find_arg_reference_data(hub, arg_bind_expr: str):
    """
    Resolve ${cloud:state:attribute_path} expressions to a value used in jinja using the hub's RUNNING dictionary
    """

    state_data = None

    run_name = hub.idem.RUN_NAME

    arg_bind_arr = arg_bind_expr.split(":")

    if len(arg_bind_arr) < 2:
        hub.log.debug(
            f" arg-bind expression `{arg_bind_expr}` doesn't comply with standards. Expected format is "
            f"$'{'resource_state:resource_name:attribute-path'}'  "
        )
        return state_data, False

    state_id = arg_bind_arr[1]

    attribute_path = None

    if len(arg_bind_arr) > 2:
        """
        From the arg-bind template - ${cloud:state:[0]:attribute:[1]} , get the attribute path [0]:attribute:[1]
        which will be used to resolve the correct value from the new_state.
        """
        attribute_path = arg_bind_arr[2:]

    run_data = hub.idem.RUNS.get(run_name, None)
    low_data = None
    if run_data:
        low_data = run_data.get("low", None)

    tag = None
    if low_data:
        for low in low_data:
            if "__id__" in low and low["__id__"] == state_id:
                chunk = {
                    "__id__": state_id,
                    "name": low.get("name"),
                    "state": low.get("state"),
                    "fun": low.get("fun"),
                }
                tag = hub.idem.tools.gen_chunk_func_tag(chunk)
                break

    arg_bind_template = "${" + arg_bind_expr[0] + "}"

    if not tag:
        hub.log.debug(
            f"Could not parse `{arg_bind_expr}` in jinja. The data for arg_binding reference `{arg_bind_template}` "
            f"could not be found on the hub. "
        )
        return state_data, False

    if run_data:
        executed_states = run_data.get("running", None)
        if executed_states is not None and tag in executed_states:
            state_data = executed_states.get(tag).get("new_state", None)
            if state_data and attribute_path:
                state_data = hub.tool.idem.arg_bind_utils.parse_dict_and_list(
                    state_id, state_data, attribute_path
                )

    return state_data, True


def parse_index(hub, key_to_parse):
    """
    Parse indexes of key. For example, test[0][1][*] will return "test" as parsed key and [0,1,*] as parsed indexes.
    """
    # Handle the case when dictionary references in the key definition are escaped, ex test[//0][//1]
    indexes = re.findall(r"\[[^\]\\]*]", key_to_parse)
    if indexes:
        parsed_indexes = []
        for index in indexes:
            parsed_index = index[1 : len(index) - 1]
            if not parsed_index.isdigit() and parsed_index != "*":
                raise ValueError(
                    f'Cannot parse argument value for key "{key_to_parse}" for index "{parsed_index}", because "{parsed_index}" is not supported".'
                )
            parsed_indexes.append(parsed_index)

        return key_to_parse[0 : key_to_parse.index("[")], parsed_indexes

    return key_to_parse, None


def get_chunk_with_index(hub, chunk, indexes, arg_key):
    if indexes:
        index = indexes[0]
        if index.isdigit():
            index = int(index)
            if not isinstance(chunk, list) or len(chunk) < index + 1:
                raise ValueError(
                    f'Cannot parse argument value for key "{arg_key}" and index "{index}", '
                    f'because argument value is not a list or it does not include element with index "{index}".'
                )
            chunk = chunk[index]
            return get_chunk_with_index(hub, chunk, indexes[1:], arg_key)
        elif index == "*":
            if not isinstance(chunk, list):
                raise ValueError(
                    f'Cannot parse argument value for key "{arg_key}" for index "{index}", '
                    f"because argument key is not a list."
                )
            chunks = [
                get_chunk_with_index(hub, item, indexes[1:], arg_key) for item in chunk
            ]
            return chunks
        else:
            raise ValueError(
                f'Cannot parse argument value for key "{arg_key}" for index "{index}", because "{index}" is not supported".'
            )

    return chunk



import fnmatch


def apply(self, high):
    """
    Read in the __exclude__ list and remove all excluded objects from the
    high data
    """
    if "__exclude__" not in high:
        return high
    ex_sls = set()
    ex_id = set()
    exclude = high.pop("__exclude__")
    for exc in exclude:
        if isinstance(exc, str):
            # The exclude statement is a string, assume it is an sls
            ex_sls.add(exc)
        if isinstance(exc, dict):
            # Explicitly declared exclude
            if len(exc) != 1:
                continue
            key = next(exc.keys())
            if key == "sls":
                ex_sls.add(exc["sls"])
            elif key == "id":
                ex_id.add(exc["id"])
    # Now the excludes have been simplified, use them
    if ex_sls:
        # There are sls excludes, find the associated ids
        for name, body in high.items():
            if name.startswith("__"):
                continue
            sls = body.get("__sls__", "")
            if not sls:
                continue
            for ex_ in ex_sls:
                if fnmatch.fnmatch(sls, ex_):
                    ex_id.add(name)
    for id_ in ex_id:
        if id_ in high:
            high.pop(id_)
    return high



from typing import Any
from typing import Awaitable
from typing import Dict

from idem.exec.init import ExecReturn


def format_return(hub, ret, ref: str, kwargs: Dict[str, Any]) -> ExecReturn:
    """
    Format the return of an exec module to be an ExecReturn class

    :param hub:
    :param ret: The data returned from the exec module
    :param ref: The reference to the exec module function on the hub
    :param kwargs: The arguments that were passed to the exec module
    """
    if isinstance(ret, ExecReturn):
        return ret

    # Get the acct metadata from ctx if it exists
    acct_details = None
    ctx = kwargs.get("ctx", {})
    if isinstance(ctx, dict):
        acct_details = ctx.get("acct_details")

    # Publish an event with the return data from the function
    hub.idem.event.put_nowait(
        body=ret,
        profile="idem-exec",
        tags={"ref": ref, "type": "exec-post", "acct_details": acct_details},
    )

    try:
        return ExecReturn(
            **ret,
            ref=ref,
        )
    except TypeError:
        raise TypeError(
            f"Exec module '{ref}' did not return a dictionary: "
            "\n{'result': True|False, 'comment': Any, 'ret': Any}"
        )


async def format_async_return(
    hub, ret: Awaitable, ref: str, kwargs: Dict[str, Any]
) -> ExecReturn:
    """
    Format the return of an async exec module to be an ExecReturn class.

    :param hub:
    :param ret: The data returned from the exec module
    :param ref: The reference to the exec module function on the hub
    :param kwargs: The arguments that were passed to the exec module
    """
    ret = await hub.pop.loop.unwrap(ret)
    return hub.tool.idem.exec.format_return(ret=ret, ref=ref, kwargs=kwargs)



from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    If changes are made then run the configured post command
    """
    ret = {}
    if reqret["ret"]["changes"]:
        func = getattr(hub, f'states.{chunk["state"]}.{condition}')
        if func:
            ret["post"] = func
    return ret



import re

# Some match examples
#  ?? params.get('rg_name').get('your_rg', 'default') ?? ^^rg_name^^.~~your_rg ??
#  ?? params['locations'][0] ?? ^^locations^^[0] ??
#  ?? params['locations'][3].get('xstate') ?? ^^locations^^[3].~~xstate ??
regex = re.compile(r"[?][?] (.+?(?= [?][?])) [?][?] .+?(?= [?][?]) [?][?]")


def _remap(hub, value):
    value = hub.tool.parameter.split_and_decode_base64(value)
    return regex.sub(r"{{ \1 }}", value)


def _remap_recursively(hub, var):
    if isinstance(var, dict):
        result = {}
        for key, value in var.items():
            result[_remap(hub, key)] = _remap_recursively(hub, value)
        return result
    elif isinstance(var, list):
        result = []
        for item in var:
            result.append(_remap_recursively(hub, item))
        return result
    elif isinstance(var, str):
        return _remap(hub, var)
    else:
        return var


def _remap_run_data(hub, run_data, keys):
    for key in keys:
        run_data[key] = _remap_recursively(hub, run_data[key])


def stage(hub, name):
    _remap_run_data(
        hub, hub.idem.RUNS[name], ["high", "low", "meta", "parameters", "warnings"]
    )



def high(hub, high):
    """
    Verify that the high data is viable and follows the data structure
    """
    errors = []
    if not isinstance(high, dict):
        errors.append("High data is not a dictionary and is invalid")
    reqs = {}
    for id_, body in high.items():
        if id_.startswith("__"):
            continue
        if not isinstance(id_, str):
            errors.append(
                f'ID "{id_}" in SLS "{body["__sls__"]}" is not formed as a string, but '
                f"is a {type(id_).__name__}"
            )
        if not isinstance(body, dict):
            err = f"The type {id_} in {body} is not formatted as a dictionary"
            errors.append(err)
            continue
        for state in body:
            if state.startswith("__"):
                continue
            if not isinstance(body[state], list):
                errors.append(
                    f'State "{id_}" in SLS "{body["__sls__"]}" is not formed as a list'
                )
            else:
                fun = 0
                if "." in state:
                    fun += 1
                for arg in body[state]:
                    if isinstance(arg, str):
                        fun += 1
                        if " " in arg.strip():
                            errors.append(
                                f'The function "{arg}" in state '
                                f'"{id_}" in SLS "{body["__sls__"]}" has '
                                "whitespace, a function with whitespace is "
                                "not supported, perhaps this is an argument "
                                'that is missing a ":"'
                            )
                    elif isinstance(arg, dict):
                        # The arg is a dict, if the arg is require or
                        # watch, it must be a list.
                        #
                        # Add the requires to the reqs dict and check them
                        # all for recursive requisites.
                        argfirst = next(iter(arg))
                        if argfirst in (
                            "require",
                            "watch",
                            "prereq",
                            "onchanges",
                            "arg_bind",
                        ):
                            if not isinstance(arg[argfirst], list):
                                errors.append(
                                    f"The {argfirst}"
                                    f' statement in state "{id_}" in SLS "{body["__sls__"]}" '
                                    "needs to be formed as a list"
                                )
                            # It is a list, verify that the members of the
                            # list are all single key dicts.
                            else:
                                reqs[id_] = {"state": state}
                                for req in arg[argfirst]:
                                    if isinstance(req, str):
                                        req = {"id": req}
                                    if not isinstance(req, dict):
                                        err = (
                                            f"Requisite declaration {req}"
                                            f' in SLS {body["__sls__"]} is not formed as a'
                                            " single key dictionary"
                                        )
                                        errors.append(err)
                                        continue
                                    req_key = next(iter(req))
                                    req_val = req[req_key]
                                    if "." in req_key:
                                        errors.append(
                                            f'Invalid requisite type "{req_key}" '
                                            f'in state "{id_}", in SLS '
                                            f'"{body["__sls__"]}". Requisite types must '
                                            "not contain dots, did you "
                                            f'mean "{req_key[:req_key.find(".")]}"?'
                                        )
                                    if not hub.idem.tools.ishashable(req_val):
                                        errors.append(
                                            f'Illegal requisite "{str(req_val)}", '
                                            f'is SLS {body["__sls__"]}\n'
                                        )
                                        continue
                                    # Check for global recursive requisites
                                    reqs[id_][req_val] = req_key
                                    if req_val in reqs:
                                        if id_ in reqs[req_val]:
                                            if reqs[req_val][id_] == state:
                                                if (
                                                    reqs[req_val]["state"]
                                                    == reqs[id_][req_val]
                                                ):
                                                    err = (
                                                        "A recursive "
                                                        "requisite was found, SLS "
                                                        f'"{body["__sls__"]}" ID "{id_}" ID "{req_val}"'
                                                    )
                                                    errors.append(err)
                            # Make sure that there is only one key in the
                            # dict
                            if len(list(arg)) != 1:
                                errors.append(
                                    "Multiple dictionaries "
                                    f'defined in argument of state "{id_}" in SLS "{body["__sls__"]}"'
                                )
                if not fun:
                    if state == "require" or state == "watch":
                        continue
                    errors.append(
                        f'No function declared in state "{state}" in'
                        f' SLS "{body["__sls__"]}"'
                    )
                elif fun > 1:
                    errors.append(
                        f'Too many functions declared in state "{state}" in '
                        f'SLS "{body["__sls__"]}"'
                    )
    return high, errors



from typing import Any
from typing import Dict


async def unlock(
    hub,
    ctx,
    provider: str = None,
    profile: str = None,
    acct_data: Dict[str, Any] = None,
):
    """
    Remove the lock from the esm profile based on the provider and profile name.
    """
    # Get the acct data from the kwargs and fallback to the acct_data in the current runtime
    if acct_data is None:
        acct_data = await hub.idem.acct.data(
            acct_key=hub.OPT.acct.get("acct_key"),
            acct_file=hub.OPT.acct.get("acct_file"),
            acct_blob=hub.OPT.acct.get("acct_blob"),
        )

    # Get the ESM ctx
    # copied from https://gitlab.com/vmware/idem/idem/-/blob/master/idem/idem/managed.py#L33
    ctx = await hub.idem.acct.ctx(
        f"esm.{provider}",
        profile=profile,
        acct_data=acct_data,
    )
    hub.log.info(f"Unlocking state run on provider {provider} using profile {profile}")
    try:
        await hub.esm[provider].exit_(ctx, None, None)
    except Exception as e:
        return {
            "comment": [f"{e.__class__.__name__}: {e}"],
            "result": False,
            "ret": None,
        }
    return {"result": True, "comment": None, "ret": {}}


def version(hub):
    """
    Get the latest supported esm version from idem
    """
    return {
        "result": True,
        "comment": None,
        "ret": ".".join(str(x) for x in hub.esm.VERSION),
    }



import asyncio
import copy


def sig(hub, ctx, name, *args, **kwargs):
    ...


def pre(hub, ctx):
    """
    Before every state, fire an event with the ref and sanitized parameters
    """
    raw_kwargs = ctx.get_arguments()
    # Fire an event with kwargs without actually modifying kwargs
    # Copying the hub will cause a recursion error
    kwargs = {k: copy.copy(v) for k, v in raw_kwargs.items() if k != "hub"}
    name = kwargs.get("name", None)

    # Don't include credentials in the fired event
    kwargs.get("ctx", {}).pop("acct", None)
    acct_details = kwargs.get("ctx", {}).get("acct_details")
    # Remove 'states.' from the ref so that it looks like the sls file
    shortened_ref = ".".join(ctx.ref.split(".")[1:])

    # This state fires the event its own way
    # If there are any more exceptions we should find a more pluggable solution
    if ctx.ref.startswith("states.exec"):
        return

    hub.idem.event.put_nowait(
        profile="idem-state",
        body={name: {f"{shortened_ref}.{ctx.func.__name__}": kwargs}},
        tags={
            "ref": f"{ctx.ref}.{ctx.func.__name__}",
            "type": "state-pre",
            "acct_details": acct_details,
        },
    )


def post(hub, ctx):
    """
    Fire an event at the end of every state
    """
    kwargs = ctx.get_arguments()
    acct_details = kwargs.get("ctx", {}).get("acct_details")

    # This state fires the event its own way
    # If there are any more exceptions we should find a more pluggable solution
    if ctx.ref.startswith("states.exec"):
        return ctx.ret

    if asyncio.iscoroutine(ctx.ret):

        async def _return_coro():
            ret = await hub.pop.loop.unwrap(ctx.ret)
            await hub.idem.event.put(
                profile="idem-state",
                body=ret,
                tags={
                    "ref": f"{ctx.ref}",
                    "type": "state-post",
                    "acct_details": acct_details,
                },
            )
            return ret

        return _return_coro()
    else:
        hub.idem.event.put_nowait(
            body=ctx.ret,
            profile="idem-state",
            tags={
                "ref": f"{ctx.ref}.{ctx.func.__name__}",
                "type": "state-post",
                "acct_details": acct_details,
            },
        )
        return ctx.ret



from typing import Any
from typing import Dict
from typing import Tuple


def previous_version(hub) -> Tuple[int, int, int]:
    return 0, 0, 0


def apply(hub, esm_cache: Dict[str, Any]) -> Dict[str, Any]:
    """
    Upgrade from esm version 0.0.0 to 1.0.0
    """
    # Get the metadata from the esm file cache
    metadata = esm_cache[hub.idem.managed.ESM_METADATA_KEY]

    # Modify the esm_version in the metadata
    metadata["version"] = (1, 0, 0)

    # Apply the change to the esm file cache
    esm_cache[hub.idem.managed.ESM_METADATA_KEY] = metadata

    return esm_cache



from typing import Any
from typing import Dict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Add the "skip_esm' flag from state to ctx
    """
    try:
        mod = getattr(hub, f"states.{chunk['state']}")

        if hasattr(mod, "SKIP_ESM"):
            chunk["ctx"]["skip_esm"] = True
    except AttributeError:
        ...

    return chunk



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Invert the requisites. If invert command line parameter is specified, requisites
    need to be reverted as well since order of execution is exact reverse in case of
    present and absent function references. The following code achieves the same.
    """
    if "invert_state" not in options or not options["invert_state"]:
        hub.log.debug("Skipping invert as not enabled")
        return seq

    tag_state_map = {}
    for ind, data in seq.items():
        tag_state_map[data["tag"]] = ind

    unmets = {}
    for data in seq.values():
        if "unmet" not in data:
            continue

        tag = data["tag"]
        for unmet in data["unmet"]:
            ind = tag_state_map[unmet]
            if ind not in unmets:
                unmets[ind] = set()
            unmets[ind].add(tag)
        data["unmet"] = set()

    for ind, unmet in unmets.items():
        seq[ind]["unmet"] = unmet

    return seq



import asyncio
import copy
from typing import Dict

DEFAULT_BATCH_SIZE = 50


async def runtime(
    hub,
    name: str,
    ctx: Dict,
    seq: Dict,
    low: Dict,
    running: Dict,
    managed_state: Dict,
    progress,
):
    """
    Execute the runtime in parallel mode
    """
    inds = []
    for ind in seq:
        if seq[ind].get("unmet"):
            # Requisites are unmet, skip this one
            continue
        inds.append(ind)
    if not inds:
        # Nothing can be run, we have hit recursive requisite,
        # or we are done
        pass

    if not hub.OPT or not hub.OPT.get("idem"):
        batch_size = DEFAULT_BATCH_SIZE
    else:
        batch_size = hub.OPT.idem.get("batch_size", DEFAULT_BATCH_SIZE)

    coros = []
    for ind in inds:
        coros.append(
            hub.idem.rules.init.run(
                name,
                # every state run should use unique ctx so that ctx.acct retains correct information during whole state execution.
                copy.deepcopy(ctx),
                low,
                seq[ind],
                running,
                hub.idem.RUNS[name]["run_num"],
                managed_state,
                seq,  # need this object in recreate_on_update rule check, to find the dependents for current seq comp.
            )
        )
        hub.idem.RUNS[name]["run_num"] += 1

    if batch_size <= 0:
        # this is running all of them in parallel
        batch_size = len(coros)
    else:
        hub.log.debug(
            f"{str(batch_size)} will be used as batch size for executing {str(len(inds))} states."
        )

    # Run groups of <batch_size> of the coroutines until they all have been processed
    while coros:
        # Grab the next group of <batch_size>
        chunk = coros[:batch_size]
        # Run the coroutines in parallel with this <batch_size>, updating the progress bar as necessary
        for fut in asyncio.as_completed(chunk):
            await fut
            # Update the progress bar once per function return
            hub.tool.progress.init.update(progress)

        # Remove those <batch_size> from the main body of coroutines
        coros = coros[batch_size:]



from typing import Any
from typing import Dict
from typing import Tuple


def sig_previous_version(hub) -> Tuple[int, int, int]:
    """
    Return the ESM version that this plugin upgrades from
    """


def sig_apply(hub, esm_cache: Dict[str, Any]) -> Dict[str, Any]:
    """
    Upgrade the esm_cache from the previous version to the new version
    """



import asyncio
import copy
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

# Sleep time in seconds between re-runs
DEFAULT_RECONCILE_WAIT = 3
# Dictionary for the default static sleep time
DEFAULT_STATIC_RECONCILE_WAIT = {"static": {"wait_in_seconds": DEFAULT_RECONCILE_WAIT}}
# Dictionary keeping state's reconcile wait time in seconds
_state_to_sleep_map = {}


async def loop(
    hub,
    pending_plugin: str,
    max_pending_reruns: int,
    name: str,
    apply_kwargs: Dict[str, Any],
):
    """
    This loop attempts to apply states.
    This function returns once all the states are successful or after MAX_RERUNS_WO_CHANGE has been reached, whatever
    occurs first.
    The sleep time between each attempt will be determined by a "wait" plugin and might change between each iterations.
    Reconciliation is required if the state is "pending" as defined by the pending plugin.
    The default pending plugin defines pending state if result  is not 'True' or there are 'changes'.

    @param hub:
    @param pending_plugin: plugin name for checking if the state is pending based on the result
    @param name: name of the run
    @param apply_kwargs: possible addition arguments
    @param max_pending_reruns: the maximum number of reruns before returning False to stop reconciliation of the resource
    :return: dictionary { "re_runs_count": <number of re-runs that occurred> }
    """
    # Record the first run as following reconciliation re-runs will
    # include subset of pending states
    first_run = copy.deepcopy(hub.idem.RUNS[name]["running"])
    tag_to_reruns_wo_change_count = {}

    # Populate wait time algorithm and values for the different states
    # in this run. State has to define __reconciliation_wait__
    # with values such as:
    # { "exponential": {"wait_in_seconds": 2, "multiplier": 10} }
    # { "static": {"wait_in_seconds": 3} }
    # { "random": {"min_value": 1, "max_value": 10} }
    hub.reconcile.basic.populate_wait_times(first_run)

    # Populate old_state for pending tags
    tag_to_old_state = _populate_old_states(first_run)
    # Concatenate comments for pending tags
    tag_to_comments = hub.reconcile.basic.populate_comments(first_run)
    current_run = first_run

    count = 0
    while True:
        pending_tags = hub.reconcile.basic.get_pending_tags(
            pending_plugin=pending_plugin,
            runs=current_run,
            tag_to_reruns_wo_change_count=tag_to_reruns_wo_change_count,
            reruns_count=count,
            max_pending_reruns=max_pending_reruns,
            ctx=apply_kwargs.get("ctx"),
        )

        if not pending_tags:
            hub.reconcile.basic.update_result(
                name, first_run, current_run, tag_to_old_state, tag_to_comments
            )
            hub.log.debug(f"Reconciliation loop returns after {count} runs.")
            return {"re_runs_count": count}

        sleep_time_sec = get_max_wait_time(hub, count, pending_tags)
        hub.log.debug(f"Sleeping {sleep_time_sec} seconds for {name}")
        await asyncio.sleep(sleep_time_sec)

        count = count + 1
        hub.log.debug(f"Retry {count} for {name}")

        last_run = current_run

        # Re-run pending states
        await hub.idem.run.init.start(name, pending_tags)

        # This run will include only the pending tags
        current_run = hub.idem.RUNS[name]["running"]
        hub.reconcile.basic.populate_comments(current_run, tag_to_comments)
        hub.reconcile.basic.populate_no_changes_count(
            last_run, current_run, tag_to_reruns_wo_change_count
        )


def get_pending_tags(
    hub,
    pending_plugin: str,
    runs: Dict,
    tag_to_reruns_wo_change_count: Dict,
    reruns_count: int,
    max_pending_reruns: int,
    ctx: Dict = None,
):
    # invoke pending plugin and populate pending tags
    pending_tags = []
    for tag in runs:
        if (
            "pending_kwargs"
            in hub.reconcile.pending[pending_plugin].is_pending.signature.parameters
        ):
            pending_kwargs = {
                "ctx": ctx,
                "reruns_wo_change_count": tag_to_reruns_wo_change_count.get(tag, 0),
                "reruns_count": reruns_count,
                "max_pending_reruns": max_pending_reruns,
            }
            if hub.reconcile.pending[pending_plugin].is_pending(
                ret=runs[tag],
                state=_tag_2_state(tag),
                **pending_kwargs,
            ):
                pending_tags.append(tag)
        elif hub.reconcile.pending[pending_plugin].is_pending(
            ret=runs[tag],
            state=_tag_2_state(tag),
        ):
            pending_tags.append(tag)

    return pending_tags


def populate_wait_times(hub, runs):
    # Populate sleep times per state
    for tag in runs:
        state = _tag_2_state(tag)
        if state not in _state_to_sleep_map.keys():
            try:
                _state_to_sleep_map[state] = getattr(
                    hub.states[state],
                    "__reconcile_wait__",
                    DEFAULT_STATIC_RECONCILE_WAIT,
                )
            except Exception as e:
                hub.log.error(
                    f"Failed to retrieve sleep time for state {state}: {e.__class__.__name__}: {e}"
                )
                _state_to_sleep_map[state] = DEFAULT_STATIC_RECONCILE_WAIT


def get_max_wait_time(hub, run_count, pending_tags: List[str]):
    # Return the maximum wait time among the pending tags
    max_sleep_time = DEFAULT_RECONCILE_WAIT
    for tag in pending_tags:
        state_wait = _state_to_sleep_map.get(
            _tag_2_state(tag), DEFAULT_STATIC_RECONCILE_WAIT
        )
        wait_alg = list(state_wait.keys())[0]
        wait_val = state_wait[wait_alg]
        sleep_time = hub.reconcile.wait[wait_alg].get(**wait_val, run_count=run_count)
        if sleep_time > max_sleep_time:
            max_sleep_time = sleep_time

    return max_sleep_time


def update_result(
    hub,
    name,
    first_run,
    last_run,
    tag_to_old_state: Dict[str, dict],
    tag_to_comments: Dict[str, tuple],
):
    # Merge last_run results into first_run, recalculate changes
    # Update 'old_state' from the original run and recalculate 'changes' to reflect
    # all the changes that occurred during reconciliation:
    # the delta between original run's old_state and last_run 'new_state'
    # Update 'comment' with the cumulative comments from all runs (w/o duplicates)
    # Set that as the 'running' on the hub
    if not last_run:
        return

    first_run.update(last_run)
    for tag in last_run:
        orig_old_state = tag_to_old_state.get(tag, None)
        last_old_state = last_run[tag].get("old_state", None)
        if orig_old_state != last_old_state:
            hub.log.debug(
                f"Replacing 'old_state' for '{_tag_2_state(tag)}': {last_old_state} with {orig_old_state}"
            )

            first_run[tag]["old_state"] = orig_old_state

            # Only when replacing 'old_state' 'changes' have to be re-calculated
            first_run[tag]["changes"] = differ.deep_diff(
                last_run[tag].get("old_state") or {},
                last_run[tag].get("new_state") or {},
            )
        if tag_to_comments.get(tag):
            first_run[tag]["comment"] = tag_to_comments.get(tag)

    hub.idem.RUNS[name]["running"] = first_run


def populate_comments(
    hub, run, tag_to_comments: Dict[str, tuple] = None
) -> Dict[str, tuple]:
    # If tag_to_comments is empty populate it with the run comments.
    # Otherwise add comments to existing
    if not tag_to_comments:
        tag_to_comments = {}
    for tag in run:
        # We expect comments to be of type tuple
        if run[tag].get("comment"):
            comment = run[tag].get("comment")
            if isinstance(comment, str):
                hub.log.debug(
                    f"Expecting a comment of type tuple but got a str: {comment}"
                )
                comment = (comment,)
            elif isinstance(comment, list):
                hub.log.debug(
                    f"Expecting a comment of type tuple but got a list: {comment}"
                )
                comment = tuple(comment)
            elif not isinstance(comment, tuple):
                hub.log.warning(
                    f"Unsupported comment type: {type(comment)}, comment: {str(comment)}"
                )
                continue

            if tag_to_comments.get(tag) and not comment[0] in tag_to_comments.get(tag):
                tag_to_comments[tag] += comment
            else:
                tag_to_comments[tag] = comment

    return tag_to_comments


def populate_no_changes_count(
    hub, last_run, current_run, tag_to_reruns_wo_change_count: Dict[str, int] = None
):
    # If tag_to_comments is empty populate it with the run comments.
    # Otherwise add comments to existing
    if not last_run or not current_run:
        return
    for tag in current_run:
        if last_run.get(tag):
            if (
                last_run[tag]["result"] == current_run[tag]["result"]
                and last_run[tag]["changes"] == current_run[tag]["changes"]
            ):
                tag_to_reruns_wo_change_count[tag] = (
                    tag_to_reruns_wo_change_count.get(tag) + 1
                    if tag_to_reruns_wo_change_count.get(tag)
                    else 1
                )
            else:
                tag_to_reruns_wo_change_count[tag] = 0


def _populate_old_states(run):
    # Keep old_state per tag from the original run
    tag_to_old_state = {}
    for tag in run:
        if run[tag].get("old_state", None):
            tag_to_old_state[tag] = copy.deepcopy(run[tag]["old_state"])
    return tag_to_old_state


def _tag_2_state(tag):
    # Get state from the tag
    return tag[0 : tag.find("_|")]



# Mapping of function refs that can be inverted and the
# corresponding function refs.
inversion_mapping = {
    "absent": "present",
    "present": "absent",
}


def apply(hub, low):
    """
    This function maps function ref to its complimentary
    value. For example, absent is converted to present,
    and present to absent.
    """

    for state in low:
        if state["fun"] in inversion_mapping:
            hub.log.debug(
                f"Inverting function ref from `{state['fun']}` to `{inversion_mapping[state['fun']]}` for state {state['__id__']}"
            )
            state["fun"] = inversion_mapping[state["fun"]]

        state["order"] *= -1

    return low



from typing import Any
from typing import Dict
from typing import Tuple


def previous_version(hub) -> Tuple[int, int, int]:
    ...


def apply(hub, esm_cache: Dict[str, Any]) -> Dict[str, Any]:
    """
    Upgrade the esm_cache from the previous version to the newest version
    """
    metadata = esm_cache[hub.idem.managed.ESM_METADATA_KEY]
    esm_version = tuple(metadata.get("version", (1, 0, 0)))

    # This is the base case for following the recursive upgrade call
    if esm_version == hub.esm.VERSION:
        hub.log.info(f"ESM cache version {esm_version} is supported by idem")
        # No work to do, the versions match
        return esm_cache

    for esm_upgrade_plugin in hub.esm.upgrade:
        prev_version = esm_upgrade_plugin.previous_version()
        if prev_version == esm_version:
            hub.log.info(f"Upgrading from ESM version {prev_version} to {esm_version}")
            esm_cache = esm_upgrade_plugin.apply(esm_cache)

            # Recursively upgrade to the next version as needed
            esm_cache = hub.esm.upgrade.init.apply(esm_cache)
            break
    else:
        # The break point of the for loop wasn't reached, throw an error
        old_version_str = ".".join(str(i) for i in esm_version)
        new_version_str = ".".join(str(i) for i in hub.esm.VERSION)
        raise RuntimeError(
            f"No ESM upgrade plugin for '{old_version_str}' to '{new_version_str}'"
        )

    return esm_cache



from typing import Any
from typing import Dict

NOOP_TAG = "omit_noop_|-noop_count_|-_|-apply"


def apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Remove states that reported success without changes
    """
    ret = {}

    noop_count = 0

    for tag in data:
        state_ret = data[tag]
        result = state_ret.get("result")
        changes = state_ret.get("changes", {})

        if result and not changes:
            # If the state was successful but there were no changes, it's a no-op

            # Keep count of no-ops in a new key
            if NOOP_TAG not in ret:
                ret[NOOP_TAG] = {"count": 0, "result": True, "comment": ""}

            noop_count += 1
        else:
            # There were changes or a failure, keep it in the results
            ret[tag] = data[tag]

    if NOOP_TAG in ret:
        # Add a comment that supplies a total of passed no-op states
        ret[NOOP_TAG]["comment"] = f"Total no-op states: {noop_count}"

    return ret



import copy
from typing import Dict

import dict_tools.differ as differ
import dict_tools.update

__func_alias__ = {"profile_": "profile"}
__contracts__ = ["soft_fail"]


async def profile_(
    hub,
    ctx,
    name: str,
    provider_name: str,
    profile_name: str = None,
    source_profile: str = None,
    acct_data: Dict = None,
    **kwargs,
):
    """
    :param hub:
    :param ctx:
    :param name: The name of the profile to add to acct
    :param provider_name: The name of the provider that this profile should be used for
    :param profile_name: The name of the new profile to add, defaults to the state name
    :param source_profile: The name of the profile to
    :param acct_data: The acct_data used as a source for existing profiles
    :param kwargs: Any extra keyword arguments will be passed directly into the new profile

    Extend the profiles of the current run with information passed to this state.
    The goal is not to write your acct_file for you with automation;.
    The purpose of this state is to dynamically create credentials for things like assuming new roles
      -- which can be re-calculated every run with negligible overhead.

    .. warning::

        This state intentionally does not implement ctx.test.
        If a profile can be created then it is used for test runs same as real runs.

    .. code-block:: yaml

        state_name:
          acct.profile:
            - provider_name: test
            - profile_name: default
            - source_profile: default
            - key_1: value_1
            - key_2: value_2
    """
    result = dict(comment=[], changes=None, new_state=None, name=name, result=True)
    if "profiles" not in hub.idem.RUNS[ctx.run_name]["acct_data"]:
        hub.idem.RUNS[ctx.run_name]["acct_data"]["profiles"] = {}

    before = list(
        hub.idem.RUNS[ctx.run_name]["acct_data"]["profiles"]
        .get(provider_name, {})
        .keys()
    )

    if profile_name is None:
        profile_name = name

    # Verify that we are not overwriting an existing profile unless explicitly asked to
    if profile_name in hub.idem.RUNS[ctx.run_name]["acct_data"]["profiles"].get(
        provider_name, {}
    ):
        result["comment"] += [
            f"Overwriting '{profile_name}' under provider '{provider_name}'"
        ]

    if ctx.test:
        result["comment"] += [
            "This state does not run in test mode",
            "If a profile can be successfully created then it is used for test runs same as real runs",
        ]

    # Copy from an existing profile first if one was given, else create a new one
    if source_profile:
        # Get the acct_data from the current run
        acct_data = acct_data or hub.idem.RUNS[ctx.run_name]["acct_data"]
        new_profile = copy.deepcopy(
            acct_data["profiles"].get(provider_name, {}).get(source_profile, {})
        )
    else:
        # Create a new raw profile
        new_profile = {}

    # Overwrite values in the new profile that exist in kwargs
    new_profile.update(kwargs)
    # Prepare the new profile to be merged onto tune RUNS dictionary
    profiles = {provider_name: {profile_name: new_profile}}

    # The value after the last dot is the plugin, everything before that is the provider sub
    provider_sub = provider_name.rsplit(".", maxsplit=1)[0]

    # Run the profiles through the gather plugins and update them with any changes
    processed_profiles = await hub.acct.init.process(
        [provider_sub], profiles, hard_fail=True
    )
    dict_tools.update.update(profiles, processed_profiles)

    # Update the profiles in the RUNS structure
    dict_tools.update.update(
        hub.idem.RUNS[ctx.run_name]["acct_data"]["profiles"], profiles
    )
    # Return the only profile keys that have been changed for this state since the values contain secure information
    after = list(
        hub.idem.RUNS[ctx.run_name]["acct_data"]["profiles"]
        .get(provider_name, {})
        .keys()
    )

    # Calculate changes directly to prevent sending a new_state to ESM
    result["changes"] = differ.deep_diff({"profiles": before}, {"profiles": after})

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """
    Always skip reconciliation for this plugin
    """
    return False



def get(hub, wait_in_seconds: int = None, **kwargs):
    # return the static wait time in seconds
    #
    # Define in state:
    # __reconcile_wait__ = {"static": {"wait_in_seconds": 10}}

    if wait_in_seconds is None or wait_in_seconds <= 0:
        raise ValueError("Expecting a single numerical value: wait time in seconds.")
    return wait_in_seconds



from typing import Any
from typing import Dict


async def sig_enter(hub, ctx):
    """
    :param hub:
    :param ctx:

    Enter the context of the enforced state manager
    Only one instance of a state run will be running for the given context.
    This function enters the context and locks the run.

    The return of this function will be passed by idem to the "handle" parameter of the exit function
    """


async def sig_exit_(hub, ctx, handle, exception: Exception):
    """
    :param hub:
    :param ctx:
    :param handle: The output of the corresponding "enter" function
    :param exception: Any exception that was raised while inside the context manager or None

    Exit the context of the enforced state manager
    """


async def sig_get_state(hub, ctx) -> Dict[str, Any]:
    """
    :param hub:
    :param ctx:

    Use the information provided in ctx.acct to retrieve the enforced managed state.
    Return it as a python dictionary.
    """


async def sig_set_state(hub, ctx, state: Dict[str, Any]):
    """
    :param hub:
    :param ctx:

    Use the information provided in ctx.acct to upload/override the enforced managed state with "state".
    """



from typing import Any
from typing import Dict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Add the 'old_state' from a previous run to the ctx from the enforced state management
    """
    managed_state = hub.idem.RUNS[name]["managed_state"]

    if managed_state:
        tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
        chunk["ctx"]["old_state"] = managed_state.get(tag, None)
    else:
        chunk["ctx"]["old_state"] = None

    return chunk



"""
This file contains routines to get sls files from references
"""
import pathlib
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

SOURCE_PATTERN = re.compile(
    r"^(?P<protocol_plugin>(?P<protocol>\w+)[\+\w+]*)://(?:(?P<profile>[-.\w]+)@)?(?P<data>.+)$"
)


def parse_source(hub, source: str) -> Tuple[str, str, str, str]:
    match = SOURCE_PATTERN.match(source)
    if not match:
        raise ValueError(f"SLS source is not a valid pattern: {source}")
    protocol = match.group("protocol")
    protocol_plugin = match.group("protocol_plugin")
    acct_profile = match.group("profile")
    data = match.group("data")
    return protocol, protocol_plugin, acct_profile, data


async def ref(hub, name: str, sls: str, sources: List[str]) -> Dict[str, Any]:
    """
    Cache the given file from the named reference point

    :param hub:
    :param name: The state run name
    :param sls: an SLS location within the given sources
    :param sources: sls-sources or params-sources

    :returns A file_name/identifier and encoded yaml content

    If the sls sources uses `acct` for authentication, then a profile is specified as part of the sls_source

    .. code-block::

        [proto]://[acct_profile]@[path]
    """
    acct_data = hub.idem.RUNS[name]["acct_data"]
    original_sls = sls

    # Find the total possible number of parent directories in this source
    possible_parents = 0
    contexts = {}
    datas = {}
    protocol = None
    protocol_plugin = None
    for source in sources:
        try:
            possibly_greater = len(re.split(r"[./\\]", source.split("://")[1]))
        except IndexError:
            continue
        if possibly_greater > possible_parents:
            possible_parents = possibly_greater

            # Save the context for this source
            protocol, protocol_plugin, acct_profile, data = hub.idem.get.parse_source(
                source
            )
            if not acct_profile:
                acct_profile = "default"
            datas[source] = data
            contexts[source] = await hub.idem.acct.ctx(
                protocol_plugin,
                profile=acct_profile,
                acct_data=acct_data,
            )

    # After each failure, imply a deeper level of relative parent
    for _ in range(possible_parents):
        # Search for the sls ref in each source (in the order the sources were defined) and return the first match
        for source in sources:
            try:
                ctx = contexts[source]
                data = datas[source]
            except KeyError:
                continue

            locs = hub.idem.get.locations(sls)

            # Try each possible location until we resolve the sls ref
            for location, actual_ref in locs:
                hub.log.debug(f"Trying to find ref '{actual_ref}' in '{location}'")
                processed_location, cache_data = await hub.source[
                    protocol_plugin
                ].cache(ctx, protocol=protocol, source=data, location=location)
                if cache_data:
                    # The reference was found and contained valid data
                    hub.log.debug(f"Found ref '{actual_ref}' in '{processed_location}'")
                    return {
                        "location": processed_location,
                        "sls": actual_ref,
                        "content": cache_data,
                    }
        # Try again with a relative source
        sls = f".{sls}"
    else:
        raise LookupError(f"Could not find SLS ref '{original_sls}' in sources")


def locations(hub, sls: str) -> List[Tuple[str, str]]:
    """
    Translate an sls ref into all possible posix-style and nt-style locations
    """
    # Try both posix and windows paths as the sls may be not be on the local filesystem
    de_normalize_ref = hub.idem.sls_source.init.de_normalize(sls)
    posix_path = pathlib.PurePosixPath(de_normalize_ref)
    nt_path = pathlib.PureWindowsPath(posix_path)

    return [
        # Append ".sls" to the tried paths
        (f"{posix_path}.sls", sls),
        (f"{nt_path}.sls", sls),
        # Append "init.sls" to the tried paths
        (str(posix_path.joinpath("init.sls")), f"{sls}.init"),
        (str(nt_path.joinpath("init.sls")), f"{sls}.init"),
        # Try the raw paths
        (str(posix_path), sls),
        (str(nt_path), sls),
        # try the de-normalized sls ref
        (de_normalize_ref, de_normalize_ref),
        (f"{de_normalize_ref}.init", f"{de_normalize_ref}.init"),
        # Try the unaltered sls
        (sls, sls),
        (f"{sls}.init", f"{sls}.init"),
    ]



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Return the sequence map that should be used to execute the lowstate
    The sequence needs to identify:
    1. recursive requisites
    2. what chunks are free to run
    3. Behavior augments for the next chunk to run
    """
    ret = {}
    for ind, chunk in enumerate(low):
        tag = hub.idem.tools.gen_chunk_func_tag(chunk)
        if tag in running:
            # Already ran this one, don't add it to the sequence
            continue
        ret[ind] = {
            "chunk": chunk,
            "reqrets": [],
            "unmet": set(),
            "tag": tag,
            "errors": [],
        }
        for req, data in hub.idem.RMAP.items():
            if (
                data.get("prereq")
                or data.get("sensitive")
                or data.get("recreate_on_update")
            ):
                continue
            if req in chunk:
                for rdef in chunk[req]:
                    if not isinstance(rdef, dict):
                        continue
                    state = next(iter(rdef))
                    if isinstance(rdef[state], list):
                        name_defs = rdef[state]
                    else:
                        name_defs = [{rdef[state]: []}]

                    for name_def in name_defs:
                        if not isinstance(name_def, dict):
                            ret[ind]["errors"].append(
                                f"{name_def} should be dictionary"
                            )
                            continue
                        name = next(iter(name_def))
                        args = name_def[name]
                        r_chunks = hub.idem.tools.get_chunks(low, state, name)
                        if not r_chunks:
                            # For arg_bind there is an additional step to check in ESM. Error will be added there
                            # if we do not find requisite in ESM
                            hub.log.debug(
                                f"Requisite {req} {state}:{name} not found in current run. "
                                f"For arg_bind requisite, there is an additional step to check in ESM"
                            )
                            if req not in ["arg_bind", "require"]:
                                ret[ind]["errors"].append(
                                    f"Requisite {req} {state}:{name} not found"
                                )
                        # TODO: Can there ever be more than one chunk with the same cloud and name?
                        for r_chunk in r_chunks:
                            r_tag = hub.idem.tools.gen_chunk_func_tag(r_chunk)
                            if r_tag in running:
                                reqret = {
                                    "req": req,
                                    "name": name,
                                    "state": state,
                                    "r_tag": r_tag,
                                    "ret": running[r_tag],
                                    "chunk": "r_chunk",
                                    "args": args,
                                }
                                # it has been run, check the rules
                                ret[ind]["reqrets"].append(reqret)
                            else:
                                ret[ind]["unmet"].add(r_tag)
    return ret



from typing import Any
from typing import Dict
from typing import Tuple

import dict_tools.data
import pop.hub


async def run(
    hub,
    path: str,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    *,
    acct_file: str = None,
    acct_key: str = None,
    acct_blob: bytes = None,
    acct_profile: str = None,
    acct_data: Dict[str, Any] = None,
    rerun_data=None,
):
    if acct_profile is None:
        acct_profile = hub.acct.DEFAULT

    args = [a for a in args]

    resolved = hub.idem.ex.resolve(path)
    path = resolved["path"]
    func = resolved["func"]
    params = resolved["params"]

    ctx = {"rerun_data": rerun_data} if rerun_data else {}
    ctx = dict_tools.data.NamespaceDict(ctx)
    if "ctx" in params:
        acct_ctx = await hub.idem.acct.ctx(
            path,
            acct_file=acct_file,
            acct_key=acct_key,
            profile=acct_profile,
            acct_blob=acct_blob,
            acct_data=acct_data,
            hard_fail=True,
            validate=True,
        )
        ctx.update(acct_ctx)
    if len(ctx) > 0:
        args.insert(0, ctx)

    ret = func(*args, **kwargs)
    return await hub.pop.loop.unwrap(ret)


def resolve(hub, path: str):
    """
    Resolve the exec module path and return the function, params, and actual reference path
    """
    if not path.startswith("exec."):
        path = f"exec.{path}"

    func = hub[path]
    if isinstance(func, pop.hub.ReverseSub):
        params = func._resolve().signature.parameters
    else:
        params = func.signature.parameters
    return {"func": func, "params": params, "path": path}


async def single(hub, path: str, *args, **kwargs):
    acct_file = hub.OPT.acct.acct_file
    acct_key = hub.OPT.acct.acct_key
    acct_profile = hub.OPT.acct.get("acct_profile", hub.acct.DEFAULT)

    ret = await hub.idem.ex.run(
        path,
        args=args,
        kwargs=kwargs,
        acct_file=acct_file,
        acct_key=acct_key,
        acct_profile=acct_profile,
    )
    return ret



from typing import Dict
from typing import List

# We keep required absent states as pending, since they will not be found in ESM
# It will result in a redundant call to 'absent' which will result in 'already absent'
REQUIRE_FILTER_OUT = [{"fun": "present"}]
ARG_BIND_FILTER_IN = [{"state": "exec"}]


async def gather(hub, ctx, name: str, pending_tags: List = None):
    low = hub.idem.RUNS[name].get("low")

    # Getting 'rerun_data' from all tags.
    # 'rerun_data' is a metadata passed between re-runs.
    rerun_data_map = {}
    for tag in hub.idem.RUNS[name]["running"]:
        if hub.idem.RUNS[name]["running"][tag].get("rerun_data"):
            rerun_data_map[tag.split("_|-")[1]] = hub.idem.RUNS[name]["running"][
                tag
            ].get("rerun_data")

    # Clearing previous running data.
    hub.idem.RUNS[name]["running"] = {}
    pending_declaration_ids = [tag.split("_|-")[1] for tag in pending_tags]

    # Gather the states that are corresponding to the pending tags,
    # and the 'require' dependencies that are not resources:
    # such as time.sleep or data.write or exec.
    low_items = []
    populate_low_items(
        low,
        pending_declaration_ids,
        low_items,
        filter_out=REQUIRE_FILTER_OUT,
        filter_in=ARG_BIND_FILTER_IN,
    )

    for key, rerun_data in rerun_data_map.items():
        for item in low_items:
            if item.get("__id__") == key:
                item["rerun_data"] = rerun_data

    return low_items


def populate_low_items(
    low: List[Dict],
    reqs: List[str],
    low_items: List[Dict],
    filter_out: List = None,
    filter_in: List = None,
):
    """
    Recursively gather low items of target(s) and their pre-requisites.
    Any reqs - declaration ID might be resolved to multiple resources,
    in case there are multiple resources under the same declaration id.
    :param low:  low data
    :param reqs: list of declaration ids of requisites ('require')
    :param low_items: gathered required low items
    :param filter_out: filter to filter out low items (exclude)
    :param filter_in: filter to include low items
    """
    if not reqs:
        return
    for req in reqs:
        req_items = [item for item in low if item.get("__id__") == req]
        # Start filtering only after first iteration, where original item(s) are added
        if len(low_items) > 0 and filter_out:
            req_items = filter_out_(req_items, filters=filter_out)
        elif len(low_items) > 0 and filter_in:
            req_items = filter_in_(req_items, filters=filter_in)
        if len(req_items) < 1:
            continue
        low_items.extend(req_items)
        for req_item in req_items:
            if req_item.get("require"):
                for required in req_item.get("require"):
                    populate_low_items(
                        low,
                        list(required.values()),
                        low_items,
                        filter_out=REQUIRE_FILTER_OUT,
                    )
            if req_item.get("arg_bind"):
                req_ids = []
                for arg_bind in req_item.get("arg_bind"):
                    for key, values in arg_bind.items():
                        for value in values:
                            for req_id, val in value.items():
                                req_ids.append(req_id)
                populate_low_items(
                    low, list(req_ids), low_items, filter_in=ARG_BIND_FILTER_IN
                )


def filter_out_(low_items: List[Dict], filters: List[Dict]) -> List[Dict]:
    # Filter out required low items that are resources (present/absent)
    # as those will be retrieved from the ESM. But any others exec/sleep/script
    # required will be executed
    if not low_items or not filters:
        return low_items

    new_low_items = []
    for item in low_items:
        match = False
        for f in filters:
            for key, value in f.items():
                if item.get(key) == value:
                    match = True
        if not match:
            new_low_items.append(item)

    return new_low_items


def filter_in_(low_items: List[Dict], filters: List[Dict]) -> List[Dict]:
    # Find low items that match the filter criteria
    if not low_items or not filters:
        return low_items

    new_low_items = []
    for item in low_items:
        for f in filters:
            for key, value in f.items():
                if item.get(key) == value:
                    new_low_items.append(item)

    return new_low_items



import copy
from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the unique requisite
    Other requisites were already processed. 'Unique' is processed last to avoid circular dependencies.
    Convert 'unique' to 'unmet' dependency, to ensure sequential execution.
    Group all the states that require uniqueness for each 'unique' function.
    Find a state in the group that can be executed and mark all the other states dependent on it.
    This logic is executed after each run.
    Next state to execute is determined like this:
    1. State without unmet requirements
    2. State without dependencies on other States of the same type
    3. State with smallest dependencies chain
    """
    # Function reference to a list of tags/states
    unique_map = {}
    # Populate unique map, function and chunks that require 'unique' handling
    for ind, data in seq.items():
        if "unique" not in data["chunk"]:
            continue
        chunk = data["chunk"]
        fun_ref = f"{chunk['state']}.{chunk['fun']}"
        if fun_ref in unique_map:
            unique_map.get(fun_ref).append(data["tag"])
        else:
            unique_map = {fun_ref: [data["tag"]]}

    if len(unique_map) == 0:
        return seq

    # Map tag to data
    tag_map = {}
    # Populate tag map
    for data in seq.values():
        tag_map[data["tag"]] = data

    for fun_ref, tag_list in unique_map.items():
        if len(tag_list) <= 1:
            continue
        hub.log.debug(f"Ordering 'unique' states for '{fun_ref}'")
        # Check for a 'free' state
        tag_list_copy = copy.deepcopy(tag_list)
        next_unique_tag = _find_next_tag(tag_list_copy, tag_map)
        assert (
            next_unique_tag
        ), f"Expecting a first tag to be unique for function {fun_ref}"
        tag_list_copy.remove(next_unique_tag)
        for tag in tag_list_copy:
            tag_map[tag]["unmet"].add(next_unique_tag)

    return seq


def _find_next_tag(tag_list, tag_map):
    # Get next tag based on the following criteria:
    # 1. Tag without unmet requirements
    # 2. Tag without dependencies on other chunks of the same type
    # 3. Tag with fewest dependency chain
    if len(tag_list) == 0:
        return None

    tags = _find_independent_tags(tag_list, tag_map)
    if len(tags) == 1:
        return tags.pop()
    if not tags:
        tags = tag_list

    # Map tag to a numerical value of its dependency tree depth
    depth_map = {}
    for tag in tags:
        depth_map[tag] = _calculate_depth(tag, tag_map, 0)
    return list(_sort_tags(depth_map).keys())[0]


def _find_independent_tags(tag_list, tag_map):
    # There must be at least one tag that is not dependent on any other
    # items from the list
    if len(tag_list) == 0:
        return []
    if len(tag_list) == 1:
        return tag_list

    independent_tags = []
    for tag in tag_list:
        unmet_for_tag = tag_map[tag].get("unmet", set())
        if len(unmet_for_tag) == 0:
            independent_tags.append(tag)
        # Tag should not be dependent on others from the list
        elif len(unmet_for_tag.intersection(tag_list)) == 0:
            independent_tags.append(tag)

    return independent_tags


def _calculate_depth(tag, tag_map, depth):
    if tag_map[tag]["unmet"] is None or len(tag_map[tag]["unmet"]) == 0:
        return depth
    max = -1
    for dependent in tag_map[tag]["unmet"]:
        d = _calculate_depth(dependent, tag_map, depth + 1)
        if d > max:
            max = d
    return max


def _sort_tags(depth_map) -> Dict:
    # Sort from lower depth to lower
    return dict(sorted(depth_map.items(), key=lambda item: item[1], reverse=False))



def stage(hub, name: str):
    """
    Apply the exclude value
    """
    low = hub.idem.ccomps.low.compile(
        hub.idem.RUNS[name]["high"], hub.idem.RUNS[name]["add_low"]
    )
    hub.idem.RUNS[name]["low"] = low



import copy
from typing import Any
from typing import Dict

from dict_tools import differ


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Check to see if the recreate_on_update can be applied.
    """
    if condition != "recreate_on_update":
        return {"errors": [f'"{condition}" is not recreate_on_update requisite.']}

    if "recreate_on_update" in chunk:
        hub.log.debug(
            f"Start applying recreate_on_update requisite for the resource {chunk['__id__']}"
        )

        if not isinstance(chunk["recreate_on_update"], dict):
            return {
                "errors": [
                    f'recreate_on_update requisite should contain a dict of parameters, not {chunk["recreate_on_update"]}'
                ]
            }

        enforced_state = hub.idem.tools.get_enforced_state(chunk, managed_state)

        if (not enforced_state) or (
            enforced_state
            and chunk.get("resource_id")
            and enforced_state.get("resource_id") != chunk.get("resource_id")
        ):
            enforced_state = None
            for key, state in managed_state.items():
                if chunk.get("resource_id") == state.get("resource_id"):
                    enforced_state = state
                    break

        # if creation flow, then skip the process.
        if not bool(enforced_state):
            hub.log.debug(
                f"As the resource {chunk['__id__']} does not exist, recreate_on_update requisite will be ignored and the resource will be created."
            )
            return {}

        hub.log.debug(
            f"Resource {chunk['__id__']} exists, check if recreation is needed."
        )

        # check whether recreation is needed or not.
        needs_recreation = hub.idem.rules.recreate_on_update.is_recreation_required(
            name, chunk, enforced_state
        )
        if not needs_recreation:
            hub.log.debug(f"No changes detected for the resource {chunk['__id__']}")
            return {}

        hub.log.debug(
            f"Start processing recreation flow for the resource {chunk['__id__']}"
        )

        create_before_destroy = chunk["recreate_on_update"].get(
            "create_before_destroy", False
        )

        if create_before_destroy:
            # create new resource, update the dependent resources and then delete old resource.
            hub.log.debug(
                f"Create the new resource for {chunk['__id__']}, update the dependent resources and destroy the old resource."
            )

            # find the dependencies
            # iterate through the execution_seq object and
            # find the dependent resources for current executing chunk.
            dependent_chunks = []
            tag = hub.idem.tools.gen_chunk_func_tag(chunk)
            for ind in execution_seq:
                seq_obj = execution_seq.get(ind)
                if seq_obj.get("unmet") and tag in seq_obj.get("unmet"):
                    dependent_chunks.append(seq_obj.get("chunk"))

            # form require requisite for the dependent chunks and
            # inject this `require` requisite to deletion chunk,
            # so that absent will wait until the update operation of dependent resources is completed.
            require_req = []
            for dependent in dependent_chunks:
                require = {dependent["state"]: dependent["__id__"]}
                require_req.append(require)

            # chunk for deletion of old resource.
            delete_chunk = hub.idem.rules.recreate_on_update.get_deletion_chunk(
                name, chunk, enforced_state
            )
            delete_chunk["require"] = require_req
            hub.idem.RUNS[name].get("add_low").append(delete_chunk)

            # marking the current chunk's resource_id as None, so that
            # this will be considered as new resource creation.
            chunk["resource_id"] = None
            chunk["recreation_flow"] = True
        else:
            # destroy old resource and then create new resource
            hub.log.debug(
                f"Destroy the old resource for {chunk['__id__']} and then create the new resource."
            )

            # chunk for deletion of old resource
            delete_chunk = hub.idem.rules.recreate_on_update.get_deletion_chunk(
                name, chunk, enforced_state
            )
            delete_chunk["recreation_flow"] = True
            hub.idem.RUNS[name].get("add_low").append(delete_chunk)

            # chunk for creating new resource
            # inject `require` requisite with deletion chunk, so that
            # the creation of new resource waits until the old resource is destroyed.
            create_chunk = copy.deepcopy(chunk)
            create_chunk["__id__"] = f"{chunk['__id__']}_create_new"
            create_chunk["resource_id"] = None
            create_chunk["require"] = [{delete_chunk["state"]: delete_chunk["__id__"]}]
            create_chunk.pop("recreate_on_update")
            create_chunk["recreation_flow"] = True
            hub.idem.RUNS[name].get("add_low").append(create_chunk)

            # Mark the current chunk to be halted from current execution.
            chunk["halt_current_execution"] = True

        hub.log.debug(
            f"End processing recreation flow for the resource {chunk['__id__']}"
        )
    return {}


def get_deletion_chunk(hub, name, chunk, enforced_state):
    delete_chunk = {
        "__id__": f"{chunk['__id__']}_delete_old",
        "name": f"{chunk['__id__']}_delete_old",
        "fun": "absent",
        "state": chunk["state"],
    }
    absent_func = hub.idem.rules.init.get_func(name, delete_chunk)
    sig = absent_func.signature
    for param_name in sig.parameters:
        if param_name == "hub" or param_name == "name":
            continue
        delete_chunk[param_name] = (
            chunk.get(param_name)
            if chunk.get(param_name)
            else enforced_state.get(param_name)
        )

    return delete_chunk


def is_recreation_required(hub, name, chunk, enforced_state):
    ignore_keys = []
    ignore_keys.extend(chunk.get("ignore_changes", []))

    # we should compare only the parameters defined for present function of the resource
    # and ignore all other parameters.
    present_func = hub.idem.rules.init.get_func(name, chunk)
    ignore_params = chunk.keys() - present_func.signature.parameters
    ignore_keys.extend(list(ignore_params))

    # when the `name` is formed with `name_prefix`, we should ignore `name` for comparison
    if chunk.get("name_prefix") and chunk.get("name_prefix") in chunk["name"]:
        ignore_keys.append("name")

    desired_state = copy.deepcopy(chunk)

    # some resources need special handling for desired state,
    # so define the `pre_process_desired_state` in
    # the corresponding resource implementation.
    if callable(getattr(hub.states[chunk["state"]], "pre_process_desired_state", None)):
        desired_state = hub.states[chunk["state"]].pre_process_desired_state(
            chunk=desired_state
        )

    # compare the enforced state and desired state.
    desired_state = _handle_null(
        desired_state=desired_state, enforced_state=enforced_state
    )
    diff = differ.deep_diff(old=enforced_state, new=desired_state, ignore=ignore_keys)
    hub.log.debug(f"difference in states: {diff} for the resource {chunk['__id__']}")

    needs_recreation = False
    if diff:
        new = diff.get("new", {})
        for key in new:
            if new[key]:
                needs_recreation = True
            elif type(new[key]) is bool and enforced_state.get(key, False) != new[key]:
                needs_recreation = True

    return needs_recreation


def _handle_null(desired_state: Dict[str, Any], enforced_state: Dict[str, Any]):
    """
    Populates items in desired_state from the enforced_state if it does not exist.
    If exists and is None, then override it with the value from enforced_state.
    """
    if isinstance(enforced_state, dict):
        for key, value in enforced_state.items():
            if key in desired_state:
                desired_value = desired_state.get(key)
                if desired_value is None:
                    desired_state[key] = value
                elif isinstance(desired_value, dict):
                    desired_state[key] = _handle_null(desired_value, value)
            else:
                desired_state[key] = value

    return desired_state
